/**************************************************************************/
/*                                                                        */
/*   Program Name :  framediv                                             */
/*                                                                        */
/*      Version:  1.00  2017-03-15                                        */
/*                                                                        */
/*      "framediv" divides a frame into header and data, then data are    */
/*      output to output file.                                            */
/*                                                                        */
/*                                                                        */
/*      Copyright (c) 2017-     T.Kondo/NICT All Right Reserved           */
/*                                                                        */
/**************************************************************************/
/*

   Ver. 2017-03-14  Start development


      usage
             framediv h_bytes d_bytes [b_skip [o_file]]
       
              where  fname -- file name
			         h_bytes ------ # of bytes of header portion in a frame
			         d_bytes ------ # of bytes of data portion in a frame
                     b_skip  -----  skip first bytes (default is 0)
                     o_file   ----- output file name (deflt is framdat.dat)

*/
#define PROGRAM_NAME "framediv"
#define PROGRAM_VERSION "Ver. 2017-03-15"
#define _CRT_SECURE_NO_WARNINGS     // to supress warning messages

#define DFLT_O_FILE "framedat.dat"    // default output file name

#include <stdio.h>
#include <signal.h> 
#include <errno.h>
#include <fcntl.h>
#include <string.h>
#include <math.h>
#include <stdlib.h>
#include <time.h>
#include <sys/types.h>
#include <sys/stat.h>

#ifdef __CYGWIN__
#ifndef linux
#define linux
#endif
#endif

#ifdef linux
#include <sys/time.h>
#endif

#ifndef WIN32
#include <unistd.h>
#endif


#ifdef WIN32
#ifndef _FORWIN32SET_
#define _FORWIN32SET_
typedef unsigned int    u_int32_t;
typedef unsigned int    uint32_t;
typedef unsigned short int    u_int16_t;
typedef unsigned short int    uint16_t;
typedef unsigned char   u_int8_t;
typedef _int64 int64_t;
#define popen(_Z,_X) _popen(_Z,_X)
#define pclose(_Z) _pclose(_Z)
#endif
#endif


typedef struct INTERNAL_USE  // for internal use
{
    char fname[128];
    char o_file[128];
    int h_bytes;
    int d_bytes;
    int b_skip;
    int numb;
    int readbytes;
	int bstep;  // read step in byte (default is 1) 
	FILE *fi;
    FILE *fo;
} INTERNAL_USE_T;


void read1line(char *line)    // original is in libipvlbi.c
{
    //char cbuf[128];
    //char *pibuf=cbuf;
    int idummy; // 2011-02-09
    char c;
    /* read 1 character by 1 character till CR code coming */
    while(1)
    {
         //scanf("%c",&c);
         idummy=scanf("%c",&c);  // 2011-02-09
         if(c=='\n' || c=='\r') goto exitread;
         //*pibuf++ =c;
         *line++ =c;
     }
     exitread: 
     //*pibuf='\0';
     //strcpy(line,cbuf);
     *line='\0';
}

void pt_intuse_ini(INTERNAL_USE_T *pt)
{
/*
   initialize INTERNAL_USE_T type variable

  */
	pt->fname[0]='\0';
	pt->o_file[0]='\0';
	pt->b_skip=0;
	pt->bstep=1;
    strcpy(pt->o_file,DFLT_O_FILE);
}


int frame_divider(INTERNAL_USE_T *pt) 
{
	unsigned char *rbuf; // read buffer 
	unsigned char *obuf; // write buffer 
	int bytesread;
	int byteswrite;
	int trbytes,twbytes;
	int i, ierr;

    ierr=0;

    trbytes=0;
    twbytes=0;

    pt->numb=pt->h_bytes+pt->d_bytes;  // bytes/frame

    rbuf=(unsigned char *)malloc(pt->numb);
    obuf=rbuf+pt->h_bytes;


    if( (pt->fi  = fopen( pt->fname, "rb" )) == NULL ) {
        printf("File %s not found! \n",pt->fname);
		ierr=-1;
        goto exitend;
    }
    printf("FILE : %s opened\n",pt->fname);
	printf("\n");

    if( (pt->fo  = fopen( pt->o_file, "wb" )) == NULL ) {
        printf("File %s open error! \n",pt->o_file);
		ierr=-1;
        fclose(pt->fi);
        goto exitend;
    }
    printf("Data portion output to  %s\n",pt->o_file);
	printf("\n");

    if(pt->b_skip > 0) {  // skip 1st b_skip bytes
        fseek(pt->fi,pt->b_skip,SEEK_CUR); // move pointer by pt->b_skip
         if( ferror( pt->fi ) ){
             ierr=-1;
             goto exitend;
         }
    }

    while(1){
        bytesread =fread( rbuf, sizeof( unsigned char ), pt->numb, pt->fi );
        if( ferror( pt->fi) || bytesread==0 ){ // File EOF or read error
           i=fclose( pt->fi ) ;
           i=fclose( pt->fo ) ;
           break;
	    }
        byteswrite=fwrite( obuf, sizeof( unsigned char ), pt->d_bytes, pt->fo );
        trbytes+=bytesread;
        twbytes+=byteswrite;
    }
    printf("  Total read bytes : %d    Total write bytes: %d\n",trbytes,twbytes);
 exitend:
    return ierr;
}

int main(int argc, char *argv[])
{
	int i;
    INTERNAL_USE_T pt;
    //if(argc < 2) {
    if(argc < 4) {
       //strcpy(fname,dflt_fname);
       printf("%s %s\n",PROGRAM_NAME,PROGRAM_VERSION);
       printf("How to run: \n");
       printf("  %s fname h_bytes d_bytes  [b_skip [o_file]]\n",PROGRAM_NAME);
       printf("\n");
       printf("    where  fname    -- file name\n");
	   printf("           h_bytes ------ # of bytes of header portion in a frame\n");
	   printf("           d_bytes ------ # of bytes of data portion in a frame\n");
	   printf("           b_skip  -----  skip first bytes (default is 0)\n");
	   printf("           o_file   ----- output file name (deflt is framdat.dat)\n");
       exit(0);
    }
    // initialize pt
    pt_intuse_ini(&pt);

	if (argc>=2){
       strcpy(pt.fname,*(argv+1));
    } 
    if(argc >=3){
        pt.h_bytes=atoi(*(argv + 2 ));
    }
    if(argc >=4){
        pt.d_bytes=atoi(*(argv + 3 ));
    }
    if(argc >=5){
        pt.b_skip=atoi(*(argv + 4 ));
    }
    if(argc >=6){
        strcpy(pt.o_file,*(argv+5));
    }

    printf("************ %s  %s ************\n",PROGRAM_NAME,PROGRAM_VERSION);

	i=frame_divider(&pt);
    return 0;
}
