/**************************************************************/
/*                                                            */
/*   File name :  lib_from_K5.c                               */
/*                                                            */
/*     Library copied from K5/src etc. for VDIF utilities     */
/*                                                            */
/*      Version:   2014-06-13                                 */
/*                                                            */
/*   Copyright (c) 2014 T.Kondo/NICT All Right Reserved       */
/*                                                            */
/**************************************************************/
/*

       Library copied from already developed K5 libraries 
       for VDIF utilities
       
       
       T.Kondo 

	   Ver. 2014-06-13


*/


#include "lib_from_K5.h"

#ifdef USE_ASM

    #ifdef WIN32
        void decode_data(int *K5pos, int wd, int ilast, int pat, int adbit)  // for VC++ inline assembler
        {
        /*
                eax  --  K5pos
                ebx  --  wd
                ecx (cl)  --  adbit
                edx  --  ilast
                esi  --  loop counter
                edi  --  temporary use of wd
        
                ebp+14h --  pat

        */
                __asm{
                    mov edx, ilast
                    mov eax, K5pos
                    mov ebx, wd
                    mov cl, byte ptr [adbit]
                    //mov ch, byte ptr [pat]  //not necessary
                    xor esi,esi  
            loop_1:
                    mov edi, ebx
                    and edi, pat
                    mov [eax+esi*4], edi
                    shr ebx, cl
                    inc esi
                    cmp esi, edx
                    jnz loop_1
                }
        }
    #else  // for gcc
        void decode_data(int *K5pos, int wd, int ilast, int pat, int adbit)
        {
        /*
                K5pos --  8(%ebp)
                wd    --  12(%ebp)
                ilast --  16(%ebp)
                pat   --  20(%ebp)
                adbit --  24(%ebp)

                eax  --  K5pos
                ebx  --  wd
                ecx (cl)  --  adbit
                edx  --  ilast
                esi  --  loop counter
                edi  --  temporary use of wd
        

        */
            asm volatile (
                 "pushl %eax \n\t"     // save register used in this function
                 "pushl %ebx \n\t"
                 "pushl %ecx \n\t"
                 "pushl %edx \n\t"
                 "pushl %edi \n\t"
                 "pushl %esi \n\t"

                 "movl 16(%ebp), %edx \n\t"
                 "movl  8(%ebp), %eax \n\t"
                 "movl 12(%ebp), %ebx \n\t"

                 "movb 24(%ebp), %cl \n\t"
                 "xor %esi, %esi \n\t"  
            "loop_10: \n\t"
                 "movl %ebx, %edi \n\t"
                 "and  20(%ebp), %edi \n\t"
                 "movl %edi, (%eax,%esi,4) \n\t"
                 "shr %cl, %ebx  \n\t"
                 "inc %esi \n\t"
                 "cmp %edx, %esi \n\t"
                 "jnz loop_10 \n\t"
                 "popl %esi \n\t"    // recover saved register 
                 "popl %edi \n\t"
                 "popl %edx \n\t"
                 "popl %ecx \n\t"
                 "popl %ebx \n\t"
                 "popl %eax \n\t"
                );
        }
    #endif
#else  // normal C function
    void decode_data(int *K5pos, int wd, int ilast, int pat, int adbit)
    {
        int i,ww, wd1;
        wd1=wd;
        for(i=0; i<ilast; i++){
            ww=(int)(wd1 & pat);
            //printf("ww wd1 pat %d %d %d",ww,wd1,pat); //debug
            K5pos[i]=ww;
            wd1>>= adbit ; 
        }
    }
#endif


/******* copy from libipvlbi.c ******/

int rvsint(int i4)
{
    // reverse integer narabi
    int iw, iwork, ipick,i;
    iw=i4;
    iwork=0;
    for(i=0; i<32; i++){
        iwork=iwork << 1;
        ipick=iw & 0x1;
        iwork= ipick | (iwork & 0xFFFFFFFE);
        iw=iw >> 1;
    }
    return iwork;
}


void read1line(char *line)    // original is in libipvlbi.c
{
    //char cbuf[128];
    //char *pibuf=cbuf;
    int idummy; // 2011-02-09
    char c;
    /* read 1 character by 1 character till CR code coming */
    while(1)
    {
         //scanf("%c",&c);
         idummy=scanf("%c",&c);  // 2011-02-09
         if(c=='\n' || c=='\r') goto exitread;
         //*pibuf++ =c;
         *line++ =c;
     }
     exitread: 
     //*pibuf='\0';
     //strcpy(line,cbuf);
     *line='\0';
}


int64_t get_file_size(char *fname)
{
    /*  get file size opened */
	int result;
#ifdef WIN32                       //2006-09-20
    struct _stati64 info;    // for VC++    
    result=_stati64(fname,&info);  // for VC++
#else
    struct stat info;
    result=stat(fname,&info);
#endif
	if(result != 0) return (int64_t)(-1);
	else
    return (int64_t)info.st_size;
    // use %lld to print out whole number that has int64_t type
}

void conv2small(char *cbuf)
{
/*  
   convert upper case letter to lower case letter 
   2007-10-29
*/
    int len,n;
    len=strlen(cbuf);
    for(n=0; n<len; n++){
        if(cbuf[n] >= 65 && (unsigned char)cbuf[n] <= 90) cbuf[n]+=32;
        //  ASCII  'A'=0x41=65
        //  ASCII  'Z'=0x5A=90
        //  ASCII  'a'=0x61=97
        //  ASCII  'z'=0x7A=122
    }
}

void conv2large(char *cbuf)
{
/*  
   convert lower case letter to upper case letter 
   2007-10-29
*/
    int len,n;
    len=strlen(cbuf);
    for(n=0; n<len; n++){
        if(cbuf[n] >= 97 && cbuf[n] <= 122) cbuf[n]-=32;
        //  ASCII  'A'=0x31=65
        //  ASCII  'Z'=0x5A=90
        //  ASCII  'a'=0x61=97
        //  ASCII  'z'=0x7A=122
    }
}

int cmnd_get_p(int argc, char *argv[], char *cmnd, char *param)  // 2014-05-22
// do not care about upper case and lower case 

// get argument like -command parameter
// search "-command" in argument precisely (i.e. charcter length must be same), 
// if it is exsted, 0 is returned as
// a function value, and next argument is returned as a parameter (ascii).
// if "-command" is not exsited, -1 is returned as a function value.
// if "-command" is not accompanied by real parameter, NULL character
// is returned as param.
// 
{
    int n;
    char cbuf[80];
    char command[80];

    param[0]='\0';

    // convert to lower case character
    strcpy(command,cmnd);
    conv2small(command);
    //search command
    //printf("%s\n",command); // debug
    n=2;
    while(n <=argc) {
        strcpy(cbuf,*(argv+n-1));
        conv2small(cbuf); // convert to small characters
        if(strlen(cbuf)==strlen(command) && strncmp(cbuf,command,strlen(command))==0){  
            n++;
            if(n > argc) return 0;
            strcpy(param,*(argv+n-1));
            return 0;
        }
        n++;
    }
    return -1;
}

int cmnd_get_n2(int argc, char *argv[], char *cmnd, char *param, char *get_cmnd)
/*

 2011-04-12

 simillar to cmnd_get_n, but add actually get commnad
 from argument as a return parameter  "get_cmnd"

// do not care about upper case and lower case 

// get argument like -command parameter
// search "-command" in argument, if it is exsted, 0 is returned as
// a function value, and next argument is returned as a parameter (ascii).
// if "-command" is not exsited, -1 is returned as a function value.
// if "-command" is not accompanied by real parameter, NULL character
// is returned as param.


*/ 
{
    int n;
    char cbuf[80];
    char command[80];

    param[0]='\0';

    // convert to lower case character
    strcpy(command,cmnd);
    conv2small(command);
    //search command
    //printf("%s\n",command); // debug
    n=2;
    while(n <=argc) {
        strcpy(cbuf,*(argv+n-1));
        conv2small(cbuf); // convert to small characters
        if(strncmp(cbuf,command,strlen(command))==0){
            strcpy(get_cmnd,cbuf);
            n++;
            if(n > argc) return 0;
            strcpy(param,*(argv+n-1));
            return 0;
        }
        n++;
    }
    return -1;
}

int cmnd_get_n(int argc, char *argv[], char *cmnd, char *param)
// do not care about upper case and lower case 

// get argument like -command parameter
// search "-command" in argument, if it is exsted, 0 is returned as
// a function value, and next argument is returned as a parameter (ascii).
// if "-command" is not exsited, -1 is returned as a function value.
// if "-command" is not accompanied by real parameter, NULL character
// is returned as param.
// 
{
    int n;
    char cbuf[80];
    char command[80];

    param[0]='\0';

    // convert to lower case character
    strcpy(command,cmnd);
    conv2small(command);
    //search command
    //printf("%s\n",command); // debug
    n=2;
    while(n <=argc) {
        strcpy(cbuf,*(argv+n-1));
        conv2small(cbuf); // convert to small characters
        if(strncmp(cbuf,command,strlen(command))==0){  
            n++;
            if(n > argc) return 0;
            strcpy(param,*(argv+n-1));
            return 0;
        }
        n++;
    }
    return -1;
}

int cmnd_get(int argc, char *argv[], char *command, char *param)
// get argument like -command parameter
// search "-command" in argument, if it is exsted, 0 is returned as
// a function value, and next argument is returned as a parameter (ascii).
// if "-command" is not exsited, -1 is returned as a function value.
// if "-command" is not accompanied by real parameter, NULL character
// is returned as param.
// 
{
    int n;
    char cbuf[80];

    param[0]='\0';

    //search command
    //printf("%s\n",command); // debug
    n=2;
    while(n <=argc) {
        strcpy(cbuf,*(argv+n-1));
        if(strncmp(cbuf,command,strlen(command))==0){  
            n++;
            if(n > argc) return 0;
            strcpy(param,*(argv+n-1));
            return 0;
        }
        n++;
    }
    return -1;
}

int cmnd_get_np(int argc, char *argv[], char *cmnd, char *param)  // copied from libipvlbi.c
// do not care about upper case and lower case 
//  but # of characters should be coincide with given cmnd
//   2016-08-18

// get argument like -command parameter
// search "-command" in argument, if it is exsted, 0 is returned as
// a function value, and next argument is returned as a parameter (ascii).
// if "-command" is not exsited, -1 is returned as a function value.
// if "-command" is not accompanied by real parameter, NULL character
// is returned as param.
// 
{
    int n;
    char cbuf[80];
    char command[80];

    param[0]='\0';

    // convert to lower case character
    strcpy(command,cmnd);
    conv2small(command);
    //search command
    //printf("%s\n",command); // debug
    n=2;
    while(n <=argc) {
        strcpy(cbuf,*(argv+n-1));
        conv2small(cbuf); // convert to small characters
        if(strncmp(cbuf,command,strlen(command))==0){  
            if(strlen(cbuf)==strlen(command)){
                n++;
                if(n > argc) return 0;
                strcpy(param,*(argv+n-1));
                return 0;
            }
        }
        n++;
    }
    return -1;
}


int cmnd_get_n2para(int argc, char *argv[], char *cmnd, char *param, char *param2)   // copied from libipvlbi.c
// do not care about upper case and lower case but   get 2 parameters

// get argument like -command parameter
// search "-command" in argument, if it is exsted, 0 is returned as
// a function value, and next argument is returned as a parameter (ascii).
// if "-command" is not exsited, -1 is returned as a function value.
// if "-command" is not accompanied by real parameter, NULL character
// is returned as param.
// 
{
    int n;
    char cbuf[80];
    char command[80];

    param[0]='\0';
    param2[0]='\0';

    // convert to lower case character
    strcpy(command,cmnd);
    conv2small(command);
    //search command
    //printf("%s\n",command); // debug
    n=2;
    while(n <=argc) {
        strcpy(cbuf,*(argv+n-1));
        conv2small(cbuf); // convert to small characters
        if(strncmp(cbuf,command,strlen(command))==0){  
            n++;
            if(n > argc) return 0;
            strcpy(param,*(argv+n-1));
            n++;
            if(n > argc) return 0;
            strcpy(param2,*(argv+n-1));
            return 0;
        }
        n++;
    }
    return -1;
}

char *select_env_dflt(char *env, char *dflt)   // original is in libipvlbi.c
/******************************************************************************************/
/*     T.Kondo       2003-07-21                                                           */
/*   Select enviromental variable or dfault                                               */
/*                                                                                        */
/*    IN                                                                                  */
/*        env --  environment variable                                                    */
/*        dflt--  default                                                                 */
/*    RETURN                                                                              */
/*        ret --  real value of env if env has real value, otherwise dflt                 */
/******************************************************************************************/
{
    char *p;
    /* default apriori file directory set */
    if(getenv(env) != NULL){
       p=getenv(env);
    } else {
       p=dflt;
    }
    return p;
}



void parse_dir_and_name(char *fpath, char *dir, char *fname) //2003-03-23
{
    int len,i,k;
    char cbuf[256];  // 2007-07-17  128->256
    len=strlen(fpath);
    // converting '\' to '/' if so
    for(i=0; i<len; i++){
       if(fpath[i]=='\\'){
            cbuf[i]='/';
       } else {
            cbuf[i]=fpath[i];
       }
    }
    cbuf[len]='\0';
    k=-1;
    for (i=len-1; i>=0; i--){
        if(cbuf[i] == '/') {
            k=i;
            break;
        }
    }
    //printf("debug %d %s\n",k,cbuf);
    if(k<0){
        dir[0]='\0';
        strcpy(fname,cbuf);
    } else if (k == len-1){
        strcpy(dir,cbuf);
        fname[0]='\0';
    } else {
        strncpy(dir,cbuf,k+1);
        dir[k+1]='\0';
        strcpy(fname,cbuf+k+1);
    }
    //printf("debug %d %s\n",k,dir);
}


int substitute_env_path_name(char *path, char *fpath)
/******************************************************************************************/
/*     T.Kondo       2003-07-21                                                           */
/*                   2007-09-05                                                           */
/*   subsitute all enviroment variables in 'path' by real ones                            */
/*      ex.  $ABCD/cout.txt  or  /ad4/$ABCD/$XYZ/cout.txt                                 */
/*            ABCD and XYZ is extracted then substituted actual enviroment variable       */
/*    IN                                                                                  */
/*        path  --  given path name (may include any environment variable)                */
/*                     eg.    $ABCD/fname   or fname                                      */
/*    RETURN                                                                              */
/*        fpath   --  real path                                                           */
/*        substitute_env_path_name -- number of environment variables found               */
/******************************************************************************************/
{
    int i,k,j,len,p1,p2, nenv;
    //char cbuf[100], env[50], tdir[80];
    char cbuf[256], env[256], tdir[256];  //2007-09-05
    strcpy(fpath,path);
    strcpy(cbuf,path);
    nenv=0;
    while(1){
       env[0]='\0';
       //found $
       len=strlen(cbuf);
       for(i=0; i<len; i++){
           if(cbuf[i]=='$') {
               p1=i;  //  start pointer
               nenv++;
               break;
           }
       }
       if(i==len) goto exit1; // nothing to do
       j=0;
       for(k=i+1; k<len; k++){
          if(cbuf[k]=='/' || cbuf[k]=='\\') {
             p2=k;
             break;
          }
          env[j]=cbuf[k];
          j++;
       }
       env[j]='\0';
       tdir[0]='\0';
       if(j>0){   // get environment
          if(getenv(env) != NULL){
             strcpy(tdir,getenv(env));
             //printf("substitute_env_in_path_name:\n");
             //printf("       \'%s\' is substituted by \'%s\'\n",env,tdir);
          //} else {
             //printf("substitute_env_in_path_name: Warning!!\n");
             //printf("       Env variable \'%s\' has no value\n",env);
          }
       }
       //insert tdir
       strncpy(fpath,cbuf,p1);
       fpath[p1]='\0';
       strcat(fpath,tdir);
       strcat(fpath,cbuf+p2);
       strcpy(cbuf,fpath);
    }
   // 
exit1:;
    return nenv;
}

void get_true_path(char *path, char *dfltenv, char *dfltdir2, char *fpath)
/******************************************************************************************/
/*     T.Kondo       2003-07-21                                                           */
/*                   2007-09-05                                                           */
/*   subsitute all enviroment variables in 'path' by real ones                            */
/*      ex.  $ABCD/cout.txt  or  /ad4/$ABCD/$XYZ/cout.txt                                 */
/*            ABCD and XYZ is extracted then substituted actual enviroment variable       */
/*    IN                                                                                  */
/*        path  --  given path name (may include any environment variable)                */
/*                     eg.    $ABCD/fname   or fname                                      */
/*        dfltenv --  environment variable for default path when no path (directory       */
/*                     portion) is accompnied by 'path'.                                  */
/*                     Note: this parameter should not start with '$'.                    */
/*        dfltdir2 --  if dfltenv is null, dfltdir2 is applied instead of dfltenv.        */
/*                     This can be enviroment varialble. In this case 1st caharacter      */
/*                     should be '$'                                                      */
/*    RETURN                                                                              */
/*        fpath   --  real path                                                           */
/*                                                                                        */
/*    examples                                                                            */
/*     Case A      path = 'fname',  dfltenv = ABC,  dfltdir2 = 'xyz'                      */
/*                1st environment variable ABC is checked and it has a value (ex.abc),    */
/*                fpath will be 'abc/fname'. If ABC is null, then dfltdir2 (xyz) is       */
/*                applied (fpath will be (xyz/abc)).  Then further environment variable   */
/*                check is carried out.                                                   */
/*     Case B      path = '$TETSU/$YOKO/$POLO/fname',  dfltenv = ABC,  dfltdir2 = 'xyz'   */
/*                1st environment variable TETSU is substituted real one,                 */
/*                next environment variable YOKO is substituted real one,                 */
/*                then environment variable POLO is substituted real one.                 */
/*                                                                                        */
/******************************************************************************************/
{
    int len,nenv;
    //char dir[80],fname[80],cbuf[100];
    char dir[256],fname[256],cbuf[256];  //2007-09-05

    strcpy(fpath,path);

    /* check file name is acompanied by directory or not */
    parse_dir_and_name_fm_fpath(path, dir, fname);
    if(dir[0]=='\0'){    // then no directory is accompnied
                              //  add directory
        if(getenv(dfltenv) != NULL){
            strcpy(dir,getenv(dfltenv));
        } else {
            strcpy(dir,"./");   //current directory
            strcpy(dir,dfltdir2);   //current directory
        } 
        len=strlen(dir);
        if(dir[len-1] != '/' && dir[len-1] != '\\'){
           dir[len]='/';
           dir[len+1]='\0';
        } 
        strcpy(fpath,dir);
        strcat(fpath,fname);
    } 
    /* check file name include environment variables */
    if((nenv=substitute_env_path_name(fpath, cbuf)) > 0){
         strcpy(fpath,cbuf);
    } 
}


void k5name_make(char *mk5name, char *outdir, int channel, int group, 
                 char *k5name)
{
/*
  ftHg̍쐬jTt@C̃[
           (MK5͌Mark5t@C)
       S[h
            MK5.k5a   ---- ch1 - ch4
            MK5.k5b   ---- ch5 - ch8
            MK5.k5c   ---- ch9 - ch12
            MK5.k5d   ---- ch13 - ch16
       P[h
            MK5.k5NN --- NN͂Qchԍ 01-16
*/
    int i,len;
    //char k5dir[80],m5dir[80], m5name[20],fname[20];
    //char k5dir[80],m5dir[80], m5name[80],fname[80];  // 2006-09-27
    char k5dir[256],m5dir[256], m5name[256],fname[256];  // 2007-07-18
    char cbuf[10];
    char ag[4][2]={"a","b","c","d"};

    parse_dir_and_name(mk5name, m5dir, m5name);


    // out directory prepare
    if(outdir[0] == '\0'){
        k5dir[0]='\0';
    } else {
        // converting '\' to '/' if so
        len=strlen(outdir);
        //printf("C %d %s\n",len,outdir);
        for(i=0; i<len; i++){
            if(outdir[i]=='\\'){
                k5dir[i]='/';
            } else {
                k5dir[i]=outdir[i];
            }
        }
        k5dir[len]='\0';
        // check apri dir name accompanied by "/" or not 
        if(k5dir[len-1] != '/') {
           k5dir[len]='/';
           k5dir[len+1]='\0';
        }
    }
   // create k5name
    strcpy(fname,m5name);
    strcat(fname,".k5");
    if(channel > 0) {
        sprintf(cbuf,"-%02d",channel);
        strcat(fname,cbuf);
    } else {
        if(group > 0 && group <5) {
           strcat(fname,ag[group-1]);
        }
    }
    strcpy(k5name,k5dir);
    strcat(k5name,fname);
}

void parse_dir_and_name_fm_fpath(char *fpath, char *dir, char *fname) 
/******************************************************************************************/
/*     T.Kondo       2003-07-21                                                           */
/*                   2007-09-05                                                           */
/*   Parse fpath to directpory portion and file name portion                              */
/*                                                                                        */
/*    IN                                                                                  */
/*        fpath --  given path name                                                       */
/*    RETURN                                                                              */
/*        dir   --  directory portion                                                     */
/*                             -1 for no environment found                                */
/*        fname --  file name                                                             */
/******************************************************************************************/
{
    int len,i,k;
    //char cbuf[128];
    char cbuf[256];  //2007-09-05
    len=strlen(fpath);
    // converting '\' to '/' if so
    for(i=0; i<len; i++){
       if(fpath[i]=='\\'){
            cbuf[i]='/';
       } else {
            cbuf[i]=fpath[i];
       }
    }
    cbuf[len]='\0';
    k=-1;
    for (i=len-1; i>=0; i--){
        if(cbuf[i] == '/') {
            k=i;
            break;
        }
    }
    //printf("debug %d %s\n",k,cbuf);
    if(k<0){
        dir[0]='\0';
        strcpy(fname,cbuf);
    } else if (k == len-1){
        strcpy(dir,cbuf);
        fname[0]='\0';
    } else {
        strncpy(dir,cbuf,k+1);
        dir[k+1]='\0';
        strcpy(fname,cbuf+k+1);
    }
    //printf("debug %d %s\n",k,dir);
}


void sec2hms(int sec, int* hh, int* mm, int* ss)
{
    int j; 
    *hh=sec/3600 ;
    j=sec-*hh*3600;
    *mm=(int)(j/60) ;
    *ss=(int)(j-*mm*60) ;
}




/****** copy from libvssp32.c ******/

void k5vssp_t_ini(K5VSSP_file_t *K5)
{
/*
   initialize K5VSSP_file_t type variable

  */
    K5->aux32=K5->head32+3;
	//K5.auxbyte1st=0; // When info get from 1st header
	K5->auxbyte1st=20; // When compulsory set info

	K5->head16=(u_int16_t*)K5->head32;
	K5->head8=(unsigned char*)K5->head32;
    K5->twordread=0;  // reset word read counter  // maybe no more use
    K5->tbytesread=0;  // reset byte read counter
    K5->pos=0;        // reset current pointer position
	K5->framenum=0;   // frame number counte
	K5->headsizew=0;  // header size in 32-bit word  usually 2 for VSSP 8 for VSSP32 2009-07-31
}



int CheckK5header_with_Read(K5VSSP_file_t *S)
{
    /* 
       Read header check to get sampling parameters 
    
       if first 64bit data is not header block, read file 
       untill finding header block.

       After call, stream pointer is set to jsut next to the header
       block.

       Return
               +ve   # of bytes read till header finding
			         (if header starts current position, then ve=0)
               -1    no header found till EOF
    */
      
       int i, i32, i8;
       i32=0;
	   i8=0;  // byte counter

       S->vflag_old=S->vflag;                  // keep current valid flag
       S->dtime_vflag_old=S->dtime_vflag;  //keep time when vflag decesion made
       S->vflag=0;

       S->headsizew=2; // for K5/VSSP
       //S->headsizew=8; // for K5/VSSP32
       /* first 7 bytes (56 bit) read */

       S->bytesread = fread( S->head8, sizeof( unsigned char ), 7, S->stream );
       S->tbytesread+=S->bytesread;
       S->pos+=S->bytesread;  // pointer position is in bytes
       if( ferror( S->stream ) || S->bytesread==0 ){ // File EOF or read error
           i=fclose( S->stream ) ;
		   i8=-1;
           goto exit_loop;
	   }
       //printf("Debug %X  %X\n",S->head32,S->head32+1);  //debug

loop1:
       /* 8 bit data read */
       S->bytesread = fread( S->head8+7, sizeof( unsigned char ), 1, S->stream );
       S->tbytesread+=S->bytesread;
       S->pos+=S->bytesread;  // pointer position is in bytes
//       if( ferror( S->stream ) || S->wordread==0 ){ // File EOF or read error
       if( ferror( S->stream ) || S->bytesread==0 ){ // File EOF or read error
	       i=fclose( S->stream ) ;
		   i8=-1;
           goto exit_loop;
	   }

//entry1:
       //if ((i=CheckK5header(S)) == 0) {  // sync detection
       if ((i=CheckK5header(S)) >= 0) {  // sync detection
           goto exit_loop;
       } else if(i==-99){
           i=fclose( S->stream ) ;
		   i8=-1;
           goto exit_loop;
       } else {
//           S->head32[0]=S->head32[1];
		   for(i=0; i<7; i++){
             S->head8[i]=S->head8[i+1];
		   }
	       i8++;
           //if(i8+2 > maxbytes) return -1;
       }
       goto loop1;
exit_loop:
       if(i8 == 0) S->vflag=1;
       S->dtime_vflag=S->seconds;
       return i8;
}


int fwdNsecK5(int nsec, K5VSSP_file_t *K5) // 2008-08-08 revised
{
    /* 
       forward fast N sec then check header
       if header doesn't exist, rewind RWIND_BYTE data at first,
       then read data slowly to find out header.
       If still header doesn't exist, rewind 1sec data
       then read data slowly to find out header.

       Note: use of fsetpos() and fgetpos() was obsoleted because of
       the compatiblity ploblem of a fpos_t type variable between linux 2.6
       and previous one

    */
    int n,k;
    int ndsec, ntimes, nrestsec;
    //int nrsec;
    double dbytes;
    double bunit=2147483647.0;  // 2^31-1
    //double bunit=40000000.0;  // for check

    if(nsec < 1) return 0; // nothing to do


    ntimes=1;
    //nrsec=0;  // debug

    dbytes=((double)K5->bytesec)*((double)nsec);

    // check maximum bytes number exceed 2^31-1
    if(dbytes > bunit) {     // then devided into sub period
        ndsec= (int)(bunit/K5->bytesec);
        ntimes=nsec/ndsec;
        nrestsec=nsec-ndsec*ntimes;
        //printf("dbytes,ntimes,ndsec,nrestsec %f %d %d %d\n",
        //    dbytes,ntimes,ndsec,nrestsec); //debug 
        for(n=0; n<ntimes; n++){
            k=fwdNsecK5org(ndsec, K5);
            //nrsec+=ndsec;   // debug
        }
        k=fwdNsecK5org(nrestsec, K5);
        //nrsec+=nrestsec;    // debug
        //printf("debug nsec, nrsec %d %d\n",nsec,nrsec);

    } else {
        k=fwdNsecK5org(nsec, K5);
    }
       
    return k;
}


int fwdNsecK5org(int nsec, K5VSSP_file_t *K5)   // 2009-03-09
												// 2009-07-31
{
    /* 
       forward fast N sec then check header
       if header doesn't exist, rewind RWIND_BYTE data at first,
       then read data slowly to find out header.
       If still header doesn't exist, rewind 1sec data
       then read data slowly to find out header.

       Note: use of fsetpos() and fgetpos() was obsoleted because of
       the compatiblity ploblem of a fpos_t type variable between linux 2.6
       and previous one

	   Note:
	   Usually this function is called when streamer pointer is located
	   at just after the haeder block.  So optimization under this condition
	   was made in Ver.2009-07-31 supposing already header block length
	   is given.

    */
    int i,k;
    //int maxword, maxbytes;
    int maxbytes;
    int nbytes, nback;
    //int64_t tword, tbytes;
    int64_t tbytes;
	int64_t pos;
//	fpos_t fpos; // debug 

    if(nsec < 1) return 0; // nothing to do

    //printf("Debug Header size (W)  %d\n",K5->headsizew);

    //tword=K5->twordread;  // save current word read counter data 
    tbytes=K5->tbytesread;  // save current byte read counter data 
    pos=K5->pos;  // save current position

    k=0;
    
//	fgetpos(K5->stream,&fpos);  // debug
//	printf("Initial fwdNsecK5org fpos = %d\n",fpos);  // debug

    nbytes=K5->bytesec*nsec;    // K5->bytesec is data portion bytes (not including header)
	nbytes+=((nsec-1)*K5->headsizew*4);  // to compensate header portion  2009-07-31
	K5->pos+=nbytes;  // pointer position to be set
    if(K5->pos < K5->size){  // not yet reached to EOF
         fseek(K5->stream,nbytes,SEEK_CUR); // move pointer by K5->bytesec
         if( ferror( K5->stream ) ){
            return -1;
         }
    } else {  // exceed EOF so pinter set to EOF
         fseek(K5->stream,(int)(K5->size - pos),SEEK_CUR); // 
         if( ferror( K5->stream ) ){
            return -1;
         }
        //K5->wordread=(int)((K5->size - pos)/4); // convert to 32bit-word unit
        K5->bytesread=(int)(K5->size - pos); 
        //K5->twordread+=(int64_t)K5->wordread;
        K5->tbytesread+=(int64_t)K5->bytesread;
        //K5->pos+=K5->wordread*4;  // pointer position is in bytes
        K5->pos+=K5->bytesread;  // pointer position is in bytes
        return -1;
    }
//	fgetpos(K5->stream,&fpos);  // debug
//	printf("fwdNsecK5org fpos = %d\n",fpos);  // debug

    //K5->wordread=nbytes/4;
    //K5->twordread+=(int64_t)K5->wordread;

    //printf("Debug nsec nbytes %d %d\n",nsec,nbytes); // debug
    K5->bytesread=nbytes;
    //K5->bytesread+=(int64_t)K5->bytesread;  //2007-02-12 bug fix
    K5->tbytesread+=(int64_t)K5->bytesread;  // 2007-02-12 bug fix


    /* read header portion */
    K5->wordread = fread(K5->head32, sizeof(u_int32_t), 2, K5->stream);  // 2009-07-30 (original again)
    //K5->wordread = fread(K5->head32, sizeof(u_int32_t), K5->headsizew, K5->stream);  //2009-03-09 bug fix

    //K5->twordread+=K5->wordread;
    K5->tbytesread+=K5->wordread*4;
    K5->pos+=K5->wordread*4;  // pointer position is in bytes
    if( ferror( K5->stream ) || K5->wordread==0 ){ // EOF or File read error
	   return -1;
    }

	//fgetpos(K5->stream,&fpos);  // debug
	//printf("Middle fwdNsecK5org fpos = %d\n",fpos);  // debug

    if((i=CheckK5header(&(*K5))) < 0){  // this is not the case of header
       if(i==-99) return -1;
       /* rewind file */
       if(REWIND_BYTE == 0){
           fseek(K5->stream,-(int)(K5->pos-pos),SEEK_CUR); // set position at initial position
           //K5->twordread=tword;            // word counter retrieve
           K5->tbytesread=tbytes;            // byte counter retrieve
           K5->pos=pos;  // retrieve initial position
           if((k=CheckK5header_with_Read(&(*K5))) < 0){ // EOF or file read error
               return -1;
           }
       } else {
           //maxword=REWIND_BYTE/4;
           maxbytes=REWIND_BYTE;
           fseek(K5->stream,-(int)REWIND_BYTE,SEEK_CUR); // set position at initial position
           //K5->twordread-=REWIND_BYTE/4;            // word counter retrieve
           K5->tbytesread-=REWIND_BYTE;            // word counter retrieve
           K5->pos-=REWIND_BYTE;  // retrieve initial position
//	fgetpos(K5->stream,&fpos);  // debug
//	printf("after rewind fwdNsecK5org fpos = %d\n",fpos);  // debug
           /* 1st trial */
           //if((k=CheckK5header_with_Read2(&(*K5),maxword)) < 0){ // EOF or file read error
           if((k=CheckK5header_with_Read2(&(*K5),2*maxbytes)) < 0){ // 2009-07-31
//	fgetpos(K5->stream,&fpos);  // debug
//	printf("after Read2Error fwdNsecK5org fpos = %d\n",fpos);  // debug
               /* 2nd trial */
               if(nsec==1){
                   nback=(int)(K5->pos-pos);  // rewind to initial position
               } else {
                   nback=K5->bytesec;   // rewind just 1 sec
               }
               fseek(K5->stream,-nback,SEEK_CUR); // set position at initial position
               //K5->twordread-=nback/4;            // word counter rewind
               K5->tbytesread-=nback;            // word counter rewind
               K5->pos-=nback;  // initial position rewind
               if((k=CheckK5header_with_Read(&(*K5))) < 0){ // EOF or file read error
                   return -1;
               }
           }
//	fgetpos(K5->stream,&fpos);  // debug
//	printf("after Read2 fwdNsecK5org fpos = %d\n",fpos);  // debug
       }
       // get headder after k bytes read
    }
//	fgetpos(K5->stream,&fpos);  // debug
//	printf("final fwdNsecK5org fpos = %d\n",fpos);  // debug
       
    return k;
}


int CheckK5header_with_Read_but_no_search(K5VSSP_file_t *S)  //2010-08-03
{
    /* 
       Read header check to get sampling parameters but no search
       when header doesn't include sync pattern
    
       Although first 64bit data is not header block, do not read file 
       to find header block.

       After call, stream pointer is set to jsut next to the supposed header
       block.

       Return
               +0    header starts current position
               -1    no header found
    */
      
       int i, i32, i8;
       i32=0;
	   i8=0;  // byte counter

       S->vflag_old=S->vflag;                  // keep current valid flag
       S->dtime_vflag_old=S->dtime_vflag;  //keep time when vflag decesion made
       S->vflag=0;

       /* first 8 bytes (64 bit) read */

       S->bytesread = fread( S->head8, sizeof( unsigned char ), 8, S->stream );
       S->tbytesread+=S->bytesread;
       S->pos+=S->bytesread;  // pointer position is in bytes
       if( ferror( S->stream ) || S->bytesread==0 ){ // File EOF or read error
           i=fclose( S->stream ) ;
		   i8=-1;
           goto exit_loop;
	   }
       //printf("Debug %08X  %08X\n",S->head32[0],S->head32[1]);  
       
       if(S->headsizew > 2){   // this is VSSP32
           S->bytesread = fread( S->head8+8, sizeof( unsigned char ), 4*(S->headsizew-2), S->stream );
           S->tbytesread+=S->bytesread;
           S->pos+=S->bytesread;  // pointer position is in bytes
           if( ferror( S->stream ) || S->bytesread==0 ){ // File EOF or read error
	          i=fclose( S->stream ) ;
		      i8=-1;
              goto exit_loop;
	       }
       }
exit_loop:
       //S->seconds=(long int)(S->head32[1] & 0x1FFFFUL);
       S->seconds=(int)(S->head32[1] & 0x1FFFFUL);  // 2011-02-07
       sec2hms(S->seconds,&S->hh,&S->mm,&S->ss);


       if(i8 == 0) S->vflag=1;
       S->dtime_vflag=S->seconds;
       return i8;
}

int CheckK5header(K5VSSP_file_t *K5)
/*
    K5 header common portion (1st 64bits) check

    if 1st 64bits show K5/VSSP32 pattern,  extended header and
    AUX field are read

    2007-03-02  allow at least 1 bit error in sync pattern
    2007-10-20  revise algorythm for ALLOW_USE_SECONDS and 
                ALLOW_USE_SAMPLE_PARAM cases
    2010-08-19  include an extended format case

   VSSP Header Structure
	  
	            W0   W1   W2   W3   W4   W5   W6   W7
	   header |----|----|----|----|----|----|----|----|
	                    
              <--VSSP-->
			  <----VSSP32--->
			                 <-------VSSP32 AUX------->

      File read pointer position at call

	            W0   W1   W2   W3   W4   W5   W6   W7
	   header |----|----|----|----|----|----|----|----|
	                    
	   at call          ^
	   when return                                    ^



	[in]
	   K5->head32[0], K5->head32[1]  -- first 64 bit of header (already read)



    return 0 : K5/VSSP Header detected
	       1 : misalignment occurred in AUX field
	       2 : other error in AUX field
		  1x : 1 bit reversal detected in sync pattern but this may be header
          -1 : K5.head32 is not header
         -99 : EOF detected or File read error

      K5->ierr  -1X:1 bit error detected in sync pattern but this may be header
*/
{
    int sdat[16]={1,2,4,8,0,0,0,0,0,0,0,0,0,0,0,0};
    int fdat[16]={40,100,200,500,1000,2000,4000,8000,16000, 
                  32000,64000,128000,256000,512000,1024000,2048000}; 
    int smplbit0, sfreq0, nch;
    u_int32_t syncpat=0x8B000000UL;    // 2nd sync pattern for K5/VSSP
    u_int32_t syncpat32=0x8C000000UL;  // 2nd sync pattern for K5/VSSP32
	int nbit1, nbit2; // 2007-03-02
    int iok, iok2;
    int flag_OK_seconds,flag_OK_sample_param;

    iok=-1;

    flag_OK_seconds=1;
    flag_OK_sample_param=1;

    K5->format=0;  // suppose K5/VSSP old format

	iok2=0;
	// First 32bit sync pattern detection
    if(K5->head32[0]==0xFFFFFFFFUL) iok2=1;

    //printf("Debug %08X %08X\n",K5->head32[0],K5->head32[1]); //debug
	nbit1=0;
    nbit2=0;

#ifdef ALLOW_1BIT_REVERSE

    // allow in case of only one bit difference in FF field
	if(iok2==0){
       w1= K5->head32[0] ^ 0xFFFFFFFFUL;
	   for(n=0; n<32; n++){
	      if((w1 & 0x1UL)== 1) nbit1++;
		  w1>>= 1;
	   }
	   if(nbit1 <= 1) iok2=1;
	}
#endif
	// 2nd sync pattern detection
	if(iok2 == 1){
        if((K5->head32[1] & 0xFF000000UL) == syncpat) K5->format=1; // this is sync block
        if((K5->head32[1] & 0xFF000000UL) == syncpat32) K5->format=2; // this is sync block
        // allow in case of only one bit difference in 2nd sync field
#ifdef ALLOW_1BIT_REVERSE

	    if(K5->format==0){       // 2007-03-02
            w2 = K5->head32[1] ^ syncpat;
			w2 >>= 24;
		    for(n=0; n<8; n++){
			   if((w2 & 0x1UL)== 1) nbit2++;
			   w2>>= 1;
			}
			if(nbit2 <= 1){
				K5->format=1;
			} else {
			   nbit2=0;
               w2 = K5->head32[1] ^ syncpat32;
			   w2 >>= 24;
		       for(n=0; n<8; n++){
			      if((w2 & 0x1UL)== 1) nbit2++;
			      w2>>= 1;
			   }
			   if(nbit2 <= 1) K5->format=2;
			}
		}
#endif

	}


 
	if(K5->format > 0){ 
            iok=0;
            /* parse sampling information and get time data */ 
            K5->spatn=(int)((K5->head32[1] >> 24) & 0xFF);       
            smplbit0 = (K5->head32[1] >> 22) & 0x3;   
            K5->adbit = sdat[smplbit0];      // AD bits
            sfreq0 = ( K5->head32[1] >> 18) & 0xF;
            K5->sfreq=fdat[sfreq0];
            nch = (K5->head32[1] >> 17) & 1;
            if (nch == 1) {
                K5->numch = 4; 
            } else {
                K5->numch = 1;
            }
            //K5->seconds=(long int)(K5->head32[1] & 0x1FFFFUL);
            K5->seconds=(int)(K5->head32[1] & 0x1FFFFUL);    // 2011-02-07
            sec2hms(K5->seconds,&K5->hh,&K5->mm,&K5->ss);
			if(K5->framenum==0){
				K5->sfreq1st=K5->sfreq;
				K5->adbit1st=K5->adbit;
				K5->numch1st=K5->numch;
				K5->seconds1st=K5->seconds;
				K5->secold=K5->seconds;
			}
			K5->framenum++;
	}

#ifdef ALLOW_USE_SECONDS
	if(K5->framenum > 1 && K5->format > 0){
        flag_OK_seconds=0;
		if((K5->seconds - K5->secold == 1) ||
			(K5->seconds - K5->secold == -86399)){   // OK
			K5->secold=K5->seconds;
            flag_OK_seconds=1;
		} 
    }
#endif

#ifdef ALLOW_USE_SAMPLE_PARAM
	if(K5->framenum > 1 && K5->format > 0){
        flag_OK_sample_param=0;
		if(K5->sfreq==K5->sfreq1st &&
			K5->adbit==K5->adbit1st &&
            K5->numch==K5->numch1st){  // OK
            flag_OK_sample_param=1;
		}
	}
#endif

    if(flag_OK_seconds == 0 && flag_OK_sample_param== 0){
	    iok=-1;
		K5->format=0;
		K5->framenum--;
    }

    if(K5->format==2){  // VSSP32 // read further header
            K5->headsizew=8;
            //if(header_aux_read(K5)<0) iok=-99;
			iok=header_aux_read(K5);
            /* add on 2010-08-19 */
            if(K5->aux_fmt==21){ // overwrite information
 			    if(K5->framenum==1){
				   K5->sfreq1st=K5->sfreq;
				   K5->numch1st=K5->numch;
                }
			}
           /* add end */
	}
	if(iok >= 0) {
		if(nbit1+nbit2 > 0){
			iok+=10;
			if(K5->ierr <= 0) K5->ierr-=10;
		}
	}

    return iok;
}


int CheckK5header_with_Read2(K5VSSP_file_t *S, int maxbytes)
{
    /* 
       Read header check to get sampling parameters 
    
       if first 64bit data is not header block, read file 
       untill finding header block.

       After call, stream pointer is set to jsut next to the header
       block.

       Return
               +ve   # of bytes read till header finding
			         (if header starts current position, then ve=0)
               -1    no header found till EOF
    */
      
       int i, i32, i8;
       i32=0;
	   i8=0;  // byte counter

       S->vflag_old=S->vflag;                  // keep current valid flag
       S->dtime_vflag_old=S->dtime_vflag;  //keep time when vflag decesion made
       S->vflag=0;

       S->headsizew=2; // for K5/VSSP
       //S->headsizew=8; // for K5/VSSP32

       /* first 7 bytes (56 bit) read */

       S->bytesread = fread( S->head8, sizeof( unsigned char ), 7, S->stream );
       S->tbytesread+=S->bytesread;
       S->pos+=S->bytesread;  // pointer position is in bytes
       if( ferror( S->stream ) || S->bytesread==0 ){ // File EOF or read error
           i=fclose( S->stream ) ;
		   i8=-1;
           goto exit_loop;
	   }
       //printf("%X  %X\n",S->head32,S->head32+1);  //debug

loop1:
       /* 8 bit data read */
       S->bytesread = fread( S->head8+7, sizeof( unsigned char ), 1, S->stream );
       S->tbytesread+=S->bytesread;
       S->pos+=S->bytesread;  // pointer position is in bytes
//       if( ferror( S->stream ) || S->wordread==0 ){ // File EOF or read error
       if( ferror( S->stream ) || S->bytesread==0 ){ // File EOF or read error
	       i=fclose( S->stream ) ;
		   i8=-1;
           goto exit_loop;
	   }

//entry1:
       //if ((i=CheckK5header(S)) == 0) {  // sync detection
       if ((i=CheckK5header(S)) >= 0) {  // sync detection
           goto exit_loop;
       } else if(i==-99){
           i=fclose( S->stream ) ;
		   i8=-1;
           goto exit_loop;
       } else {
//           S->head32[0]=S->head32[1];
		   for(i=0; i<7; i++){
             S->head8[i]=S->head8[i+1];
		   }
	       i8++;
           if(i8+2 > maxbytes) return -1;
       }
       goto loop1;
exit_loop:
       if(i8 == 0) S->vflag=1;
       S->dtime_vflag=S->seconds;
       return i8;
}

int header_aux_read(K5VSSP_file_t *K5)
{
/*
  read Aux field in K5/VSSP32 header
  assumption : pointer is located just next to 64bit common header
    
  Remarks:  If K5->auxbyte1st is 0 when call this function first time,
             auxbyte information got from 1st header will be 
			 sbustituted to K5->auxbyte1st
  return  
          0: normal end
		 -99: EOF or FILE read
		  1: misalignment occurred in AUX field
		  2: other error in AUX field

      K5->ierr  0: normal end
		        -1: misalignment occurred in AUX field
		        -2: other error in AUX field

        2010-08-19    include an extended format case

  */

       int i, i32, n;
       int iw;

       i32=0;

	   K5->ierr=0;
       /* next 32bit data read */
       K5->wordread = fread( K5->head32+2, sizeof( u_int32_t ), 1, K5->stream );
       //K5->twordread+=K5->wordread;
       K5->tbytesread+=K5->wordread*4;
       K5->pos+=K5->wordread*4;  // pointer position is in bytes
       if( ferror( K5->stream ) || K5->wordread==0 ){ // File EOF or read error
	       i=fclose( K5->stream ) ;
	       i32=-99;
           goto exit_here;
       }
       // parse header into each item
       K5->eflg=(int)((K5->head32[2] >> 15) & 0x1); // error flag  2007-02-07
       //K5->year=2000+(int)((K5->head32[2] >> 9) & 0x7F); // year
       K5->year=2000+(int)((K5->head32[2] >> 9) & 0x3F); // year  2007-02-07
       K5->totalday=(int)(K5->head32[2] & 0x1FF);
       K5->version1=(int)((K5->head32[2] >> 28) & 0xF);
       K5->version2=(int)((K5->head32[2] >> 24) & 0xF);
       K5->auxbyte =(int)((K5->head32[2] >> 16) & 0xFF);
       //printf("year = %d auxbyte = %d\n",K5->year,K5->auxbyte);
	   if(K5->auxbyte1st == 0) K5->auxbyte1st=K5->auxbyte;
	   if(K5->auxbyte != K5->auxbyte1st){
		   //printf("\n # of aux bytes (%d) is different from default value (%d)!\n",
           //       K5->auxbyte,K5->auxbyte1st );
#ifdef ALLOW_AUX_MISALIGN
		   // check AUX field misalignment
	       if(K5->framenum > 1){
               K5->auxbyte =(int)(K5->head32[2] & 0xFF);
			   if(K5->auxbyte == K5->auxbyte1st){  // AUX field misalignement may be occurre
                   // re-parse header into each item
                   K5->eflg=(int)((K5->head32[2] >> 15) & 0x1); 
                   K5->version1=(int)((K5->head32[2] >> 12) & 0xF);
                   K5->version2=(int)((K5->head32[2] >> 8) & 0xF);
				   i32=1;
				   K5->ierr=-1;
			   }
		   }
#endif
           //printf("Debg AUX error %d %d\n",K5->auxbyte,K5->auxbyte1st);
		   if(i32 !=1){
              K5->auxbyte=K5->auxbyte1st;
		      i32=2;
		      K5->ierr=-2; // means other error in AUX field
		   }
	   }


	   // further header read
       //for(n=0; n<K5->auxbyte/4; n++){
       for(n=0; n<K5->auxbyte1st/4; n++){  // 2007-10-19 to make robust
           K5->wordread = fread( K5->head32+3+n, sizeof( u_int32_t ), 1, K5->stream );
           //K5->twordread+=K5->wordread;
           K5->tbytesread+=K5->wordread*4;
           K5->pos+=K5->wordread*4;  // pointer position is in bytes
		   //printf("Debug word read %d  %d  %d\n",n, K5->wordread,K5->ierr);
           if( ferror( K5->stream ) || K5->wordread==0 ){ // File EOF or read error
	           i=fclose( K5->stream ) ;
		       i32=-99;
               goto exit_here;
	       }

		   if(i32==1){
              K5->eflg=(int)((K5->head32[2+K5->auxbyte/4] >> 31) & 0x1); 
              K5->year=2000+(int)((K5->head32[2+K5->auxbyte/4] >> 25) & 0x3F); 
              K5->totalday=(int)((K5->head32[2+K5->auxbyte/4] >> 16) & 0x1FF);
		   }
       }
       /* add on 2010-08-19 */
       K5->aux_fmt =(int)(K5->head32[3] & 0xFF);
       if(K5->aux_fmt==21){  // extended format
           /* overwrite channel and sampling frequency info */
           K5->numch=1<< (int)((K5->head32[3]>>16) & 0x7);
           iw=(int)((K5->head32[3]>>19) & 0x1FFF)*1000;  
           if(iw !=0)  K5->sfreq=iw;
       }
       /* add end */


exit_here:
       //printf("Debug %d %d\n",  i32,K5->auxbyte1st);
       return i32;

}

void vssp_aux_parse(vsspaux_t *vsspaux)
{
/*  parse AUX field data 
    [in]
      vsspaux->size  ----  AUX filed size in bytes
      vsspaux->bdat  ----  AUX field
    [return]
      vsspaux->fmt;  ----  AUX field format
      vsspaux->filter;  ----  FILTER parameter ( when fmt !=0)
      vsspaux->sid2; ----  2 char Station ID ( when fmt=1)
      vsspaux->station8;  -- station name (8 char)
      vsspaux->hostname8; -- Host PC name (8 char)



*/

    int n;
    /* fill out with space code */
    strcpy(vsspaux->sid," ");
    strcpy(vsspaux->sid2,"  ");
    strcpy(vsspaux->station8,"        ");
    strcpy(vsspaux->hostname8,"        ");
    
    /* format field get */
    vsspaux->fmt=(int)vsspaux->bdat[0];
    /* filter field get */
    vsspaux->filter=(int)vsspaux->bdat[1];
    if(vsspaux->fmt==1){
       /* station field get */
       vsspaux->sid[0]=vsspaux->bdat[1];
       vsspaux->sid[1]='\0';
       vsspaux->sid2[0]=vsspaux->bdat[2];
       vsspaux->sid2[1]=vsspaux->bdat[3];
       vsspaux->sid2[2]='\0';
       for(n=0; n<8; n++) vsspaux->station8[n]=vsspaux->bdat[4+n];
       vsspaux->station8[8]='\0';
       /* hostname field set */
       for(n=0; n<8; n++) vsspaux->hostname8[n]=vsspaux->bdat[12+n];
       vsspaux->hostname8[8]='\0';
    } else if(vsspaux->fmt==2){
       /* hostname field set */
       for(n=0; n<8; n++) vsspaux->hostname8[n]=vsspaux->bdat[12+n];
       vsspaux->hostname8[8]='\0';
    }
}
