/**************************************************************/
/*                                                            */
/*   File name :  libspeplot_v.h                              */
/*                                                            */
/*      Version:   2018-02-11                                 */
/*                                                            */
/*   Copyright (c) 2018- T.Kondo      All Right Reserved      */
/*                                                            */
/**************************************************************/
/*

2018-02-11
  "wpows" is added asa member of speanav_param_t type variable
2007-11-13
  divide libspeplot.h into libspeplot.h and libspeplot.c
2007-11-04  
  improve appearance in case of GNUPLOT postscript
2007-11-02  
  support GNUPLOT
2006-01-31
  libspeplot.h includes functions used for spectrum plot.

*/
#ifndef _LIBSPEPLOT_
#define _LIBSPEPLOT_

#include "chead.h"
#include "g_head.h"    // header for graphics
#include "libvdif.h"

#ifdef WIN32
extern tempcounter;  // counter for tempfile manage
#endif

typedef struct speanav_param   // speana parameter type variable
{
    char fname[128]; // data file name
    char ftemp[128]; // temporary file name
    char g_device[30]; // current graphic device name 
    char g_display[30]; // graphic device name for display in case of GNUPLOT
    char g_ps_device[30]; // graphic device name for POSTSCRIPT
    char ps_fname[128]; // Postscript out file name
    char comment[60];
    char title[60];
    char lab1[128];
    char lab2[128];
    double sfreq;
	int frames_per_sec;  // # frames per sec
	int frame_len64;    // Data fram length (including header) in units of 8 bytes (must be multiple of 8 bytes) 
    int adbit;
    int numch;
    int nfft;
    int amode;
    int pmode;
    int mode2;
    int mode;
    int time1st[3]; // hh,mm,ss
    float *spows, *xdata, *ydata, *fdat;
    float *wpows;  // 2018-02-11
    double sekibun1;  // given unit integration period
    double sekibun;  // given integration period
    double tinteg;  // actual integration period
    double soffset;  // start offset (in sec)
    double f1khz;  
    double f2khz;
    double rfmhz;  // 2018-04-09 base band RF frequency (MHz) (used in case of flag_velocity=1) -VE means LSB
    double refmhz; // 2018-04-09 RF frequency where set velocity = 0  (used in case of flag_velocity=1)
    int flag_no_info;        // 2018-04-09  suppres display of file name anfnd processing date
    int flag_velocity;       // 2018-04-09  display velocity instead of frequency
    int flag_no_title;       // 2018-04-10  title display flag 1:no display 0: display title
    int mode_title;       // 2018-04-10 title display mode 0: full 1: omit time info: 2: omit both sampling and time
    int flag_complex;     // 2018-04-26   1: I/Q data 0: real data (normal data)
    int flag_lsb;         // 2018-04-26   1: set LSB mode in case of I/Q data   0: suppose USB in case of I/Q data
    int mode_method;  // 2018-04-26
    FILE *gp;   // for GNUPLOT 

} speanav_param_t;

void tmpnameget(char *tmpdir, char *fpath);  // original is in libcor.c
void plotspe_v(speanav_param_t *st, float ymaxdb, float ymindb);
void plotspe2_v(speanav_param_t *st);

void plotspe_a_v(speanav_param_t *st,float ymaxdb, float ymindb);
void plotspe_a2_v(speanav_param_t *st,float *pmaxdat, 
               float *fdatmax, int nmax);
void plotautocor_v(speanav_param_t *st);

void gnline(FILE *gp,  int ndata,float *xdata, float *ydata);
void gnline2(FILE *gp,  int ndata,float *xdata, float *ydata, char *ftemp);

#endif
