/**************************************************************/
/*                                                            */
/*   File name :  libvdif.c                                   */
/*                                                            */
/*      Version:   2019-12-11                                 */
/*                                                            */
/*   Copyright (c) 2014-2019 T.Kondo      All Right Reserved  */
/*                                                            */
/**************************************************************/
/*

       Library for VDIF handling

          T.Kondo 

		      2014-04-23 start development
              2014-06-13 Ver 1.0
              2016-08-12 serious bug in "vdiftime2ymdhms" was revised
                         Time after July 1 st was wrong
              2018-02-03 "vdif_read_unit" was revised to support # of cahnnels exceedin 16 (MAX is 64)
              2018-02-09 "parse_vdif_header" is revised
              2018-07-23 add function "fwdNsecVDIF_multi"
              2018-12-12 add setting on the second function to "fwdNsecVDIF"
              2019-09-23 revise "parse_vdif_header"
              2019-10-23 add function "fwdNframeVDIF_multi"
              2019-10-31 debug function "fwdNframeVDIF_multi"
              2019-12-11 add "setpos_1st_frame" function (moved from vdifsch2m.c)
*/


#include "libvdif.h"

void parse_vdif_header(int *head8, vdif_header_t *vhead)
/*
     parse VDIF header 8x32bit words to each item
	   Ver. 2014-05-14
            2018-02-09 revised
            2019-09-23 revised
*/
{
	/* supposing little endian */

	//int i;

	//for(i=0; i<4; i++) head8[i]=rvsint(head8[i]);

    /* common portion (Word #0 - #3)*/

	vhead->invalid= head8[0] >> 31 ;
	vhead->legacy = (head8[0] >> 30) & 0x1 ;
	vhead->sec_fm_ref= head8[0] & 0x3FFFFFFF ;
	vhead->ref_epo=(head8[1] >> 24) & 0x3F ;
	vhead->frame = head8[1] & 0xFFFFFF ;
	vhead->ver = (head8[2] >> 29) & 0x7 ;
	vhead->log2_chs = (head8[2] >> 24) & 0x1F ;
	vhead->frame_len64 = head8[2] & 0xFFFFFF ;
	//vhead->d_type = head8[3] >> 31 ;
	vhead->d_type = (head8[3] >> 31) & 0x1 ;    // 2018-02-03
	vhead->bpsampl = (head8[3] >> 26) & 0x1F ;
	vhead->thread_id =  (head8[3] >> 16) & 0x3FF ;
	vhead->station_id =  head8[3] & 0xFFFF ;
	//vhead->sid[0] =  (head8[3] >> 8) & 0xFF ;
	//vhead->sid[1] =  head8[3] & 0xFF ;
	vhead->sid[1] =  (head8[3] >> 8) & 0xFF ;    // 2016-08-02
	vhead->sid[0] =  head8[3] & 0xFF ;           // 2016-08-02
	vhead->sid[2]='\0';

    /* extended portion (Word #4 - #7) */
	vhead->edv = head8[4] >> 24 ;    // Extended Data Version
	vhead->w1  = head8[4] & 0xFFFFFF ;
	vhead->w2  = head8[5] ;
	vhead->w3  = head8[6] ;
	vhead->w4  = head8[7] ;

	// check ASCII or not in sid data  (revised on 2018-02-03)
	//if(vhead->sid[0] < 0x20 || vhead->sid[0] > 0x7E) vhead->sid[0]=' '; // this is not a printable ASCII
	//if(vhead->sid[1] < 0x20 || vhead->sid[1] > 0x7E) vhead->sid[1]=' '; // this is not a printable ASCII
    if(vhead->sid[0] < 48){       // vhead->station_id is an unsigned 16-bit integer not 2 characters
        vhead->sid[0]='-';
        vhead->sid[1]='-';
    }


	/* Eextended data analysis */

	if(vhead->edv==1){    // This is NICT's EDV#1
	   vhead->uflag = (head8[4] >> 23) & 0x1; 
	   vhead->srate = head8[4] & 0x7FFFFF ;
	   vhead->sync_block = vhead->w2;
	   strncpy(vhead->das,(char *)head8+24,8);
	   vhead->das[8]='\0';
    } else if(vhead->edv==3){    // This is VLBA EDV#3   // 2018-02-09
	   vhead->uflag = (head8[4] >> 23) & 0x1;  // 2018-02-09
	   vhead->srate = head8[4] & 0x7FFFFF ;    // 2018-02-09
	   vhead->sync_block = vhead->w2;          // 2018-02-09
    } else {                                   // 2018-02-09
       vhead->uflag=0;                         // 2018-02-09
       vhead->srate=0;                         // 2018-02-09
       vhead->das[0] = '\0';                 // 2019-09-23
       vhead->sync_block = 0;                // 2019-09-23
    }

	// VDIF time convert user friendly time
	vhead->t.ref_epo=vhead->ref_epo;
	vhead->t.sec_fm_ref=vhead->sec_fm_ref;
    vdiftime2ymdhms(&vhead->t);
}


/* UTC to MJD conversion */
//long ut2mjd(double *d, int year, int month, int day,
void ut2mjd_v(timeval_vdif_t *t)
//int year, int month, int day,     // 2011-02-09
//           int hour, int min, int sec)
/*
  [in]
     year   -- year (yyyy)
     month  -- month (month)
     mday    -- day of month (dd)
     hh   -- hour (hh)
     mm    -- minute (mm)
     ss    -- second (ss)
  [return]
     mjd    -- MJD  (integer day)
     fday    -- fractional part of MJD (day)
*/
{
    int  bc, gregory, mnth, y4; 
    //long j;
    int j;   // 2011-02-09

    mnth=t->month;
    y4=t->year;
    bc = t->year <= 0;
    
    /* First day at Gregory's calendar */
    gregory = t->year > 1582 || (t->year == 1582 && t->month > 10) ||
              (t->year == 1582 && t->month == 10 && t->mday >= 15);
            
    if (mnth <= 2) {
        y4--;
        mnth += 12;
    }
    
    if (t->hh < 12) {
        j = 0;
		t->fday = 0.5;
    } else {
        j = 1;
		t->fday = -0.5;
    }
	t->fday += (t->hh*3600L + t->mm*60 + t->ss)/86400.0;
    
    j += bc ? (y4-3)/4 : y4/4;
    if (gregory) j+= 2 - y4/100 + y4/400;
	j += 1720994L + y4*365L + (mnth+1)*30 + (mnth+1)*3/5 + t->mday;
    
    /* convert to MJD */
    if (t->hh < 12) {
		t->fday -= 0.5;
    } else {
		t->fday += 0.5;
        j -= 1L;
    }
    j -= 2400000L;
	t->mjd=j;
}

/* MJD to UTC conversion */
void mjd2ut_v(timeval_vdif_t *t)
//int *year, int *month, int *day,
//           int *hour, int *min, int *sec, int j, double d)  // 2011-02-09
/*
  [in]
     mjd      -- MJD  (integer day)
     fday      -- fractional part of MJD (day)
  [return]
     year   -- year (yyyy)
     month  -- month (month)
     mday    -- day of month (dd)
     hh   -- hour (hh)
     mm    -- minute (mm)
     ss    -- second (ss)
*/
{
    int c, k, e, s; 
	int j;
	double d;

	j=t->mjd;
	d=t->fday;
    /* convert to JD */
	if( d >= 0.5) {
		d -= 0.5;
		j++;
    } else {
		d += 0.5;
    }
	j += 2400000L;


    /* JD to UTC */
    if (d >= 0.5) {
        j++;
        d -= 0.5;
    } else d += 0.5;

    if (j >= 2299161) j = j + 1 + (j-1867216)/36524 - (j-1867216)/146096;
    j += 1524;

    c = (int)((j - 122.1)/365.25); 
    k = 365L*c + c/4;
    e = (int)((j - k)/30.6001); 

	t->year = c - 4716;
	t->month = e - 1;
    if (t->month > 12) {
       t->month -= 12;
	   t->year++;
    }
	t->mday = j - k - (int)(30.6*e);

	s = (int)(d * 86400 + 0.5); 
	t->hh = s / 3600;
	t->mm = (s % 3600) / 60;
	t->ss = s % 60;
}

void vdiftime2ymdhms(timeval_vdif_t *t)
{
/*  
      convert VDIF time to normal time expression
	       T.Kondo   2014-04-23

           2016-08-12   serious bug was found   and revised
		   
*/
	//int year0, month0;
	int day0=1;
	int mjd0;
	//double d0;
	int days, secs;

    // get mjd of year start to get day of year later
	t->year=2000+(int)(t->ref_epo/2);
	t->month=1;
	t->mday=1;
	t->hh=0; t->mm=0; t->ss=0;
    ut2mjd_v(t);
	mjd0=t->mjd;
	// get mjd of 6 month epoch
	if ((t->ref_epo % 2) == 1){
		//t->month=6;
		t->month=7;  // 2016-08-12
        ut2mjd_v(t);
	}
	// convert seconds from epoch to days and seconds
	days= t->sec_fm_ref / 86400;
	secs= t->sec_fm_ref % 86400;
	// get year, month, day of month of vdif time
	t->mjd+=days;
    mjd2ut_v(t);
	// convert second in a day to hh, mm, ss
	t->hh=secs/3600;
	t->mm=(secs-t->hh*3600)/60;
	t->ss=secs % 60;
	// other parameter sets
	t->second=secs;
	t->ddd=t->mjd-mjd0+1;
	t->dtime=(double)t->ddd+(double)secs/86400.0;
	t->fday=fmod(t->dtime,1.0);
}


int check_vdif_header(FILE *fin)
{
/*
     read headder portion (32byte) from the current file pointer
	 then judge whether it is a correct VDIF header or not.
	 
     file pointer is retrieved after reading header portion

	 [return]
	        0:  may be VDIF header
		   -1:  not a VDIF header

*/
	unsigned int head[8];
	int headlen=8;
	int iread,iok;
	int64_t wpos;
    vdif_header_t vhead;

	iok=-1;
	wpos=ftell(fin);  // save current pointer position
    iread = fread(head, sizeof( unsigned int ), headlen, fin ); 
	if (iread ==0 ) { // EOF
       goto exitreturn; 
    }
    parse_vdif_header(head, &vhead);
	if(vhead.frame_len64 <= 0) goto exitreturn;
	if(vhead.t.year <= 0 || vhead.t.year > 2050) goto exitreturn;
	iok=0;
exitreturn:
    fseek(fin,wpos,SEEK_SET);
	return iok;
}

int find_Kas_sync(FILE *fin, int64_t *pos, int mode)
{
/*
     set pointer to the vdif header based on Kashima's sync pattern

   [in]
         mode --- sync search mode
		          0: limit around given 4byte word range (default)
				  1: search forward (until EOF)
				  2: search back ward (unti Begining of File)
   [return]
         pos  --- current file pointer (when header found, pointer set to the head of header)
		 find_Kas_sync    = 0  OK,    = -1 error
         
*/
    unsigned int syncpat=VDIF_KASHIMA_SYNC;    // sync pattern for VDIF Kashima format
	unsigned int ibuf;
	int offset=20;                 // offset to sync pattern from header start (in byte)
	int iread,ierr;
    int64_t wpos,wpos1,wpos2,wp;

	wpos=ftell(fin);  // save current pointer position

	ierr=-1;
	switch (mode){
		case 0:
			wpos1=wpos-HEAD_SEARCH_RANGE/2;
			wpos2=wpos1+HEAD_SEARCH_RANGE;
			wp=wpos1;
			while(wp <= wpos2){
               fseek(fin,wp,SEEK_SET);
               iread = fread( &ibuf, sizeof( unsigned int ), 1, fin ); 
               if (iread ==0 ) { // EOF
                   goto exithere; 
			   }
			   if(ibuf==syncpat) {
				   ierr=0;
				   wpos=wp-offset;
				   goto exithere;
			   }
			   wp+=4;
            }
			break;
		case 1:
			wp=wpos;
			while(1){
               fseek(fin,wp,SEEK_SET);
               iread = fread( &ibuf, sizeof( unsigned int ), 1, fin ); 
               if (iread ==0 ) { // EOF
                   goto exithere; 
			   }
			   if(ibuf==syncpat) {
				   ierr=0;
				   wpos=wp-offset;
				   goto exithere;
			   }
			   wp+=4;
            }
			break;
		case 2:
			wp=wpos;
			while(1){
               fseek(fin,wp,SEEK_SET);
               iread = fread( &ibuf, sizeof( unsigned int ), 1, fin ); 
               if (iread ==0 ) { // EOF
                   goto exithere; 
			   }
			   if(ibuf==syncpat) {
				   ierr=0;
				   wpos=wp-offset;
				   goto exithere;
			   }
			   wp-=4;
			   if(wp < 0) goto exithere;
            }
			break;

	}

exithere:
    fseek(fin,wpos,SEEK_SET);
	*pos=wpos;

	return ierr;
}


int get_real_vdif_info(vdif_file_t *vd, int mode)
{
/*
       Ver.2014-06-12

       get vdif info by reading data
	   sampling frequency is also estimated 
	   independently if header doesn't including the information or
       mode is set so
       
       VDIF file is closed at end

       in
             mode  : set information get mode
                     0:  automatic mode
                         if header includes sampling frequency, use it
                     1:  estimate compulsory by reading data for 1 sec

	   return   -99: File not found
	             -1: error occured
	              0: OK
*/

    //u_int32_t head[8];
    unsigned int syncpat=VDIF_KASHIMA_SYNC;    // sync pattern for VDIF Kashima format
	char cbuf[10]; // for operator interface (if necessary)
	int i,iread;
	//int headlen; 
	int edv;
	int iok;
	int frame_c;
	int sec_old;
	int ref_thread_id;
	int nfram_per_sec;  //  true # of frames per sec
	int nfram_per_sec_flag;  //  true # of frames per sec count OK flag
	int nthread; // true # of threads
	//int thread_id_old;
	int thread_flag;
    double pframe;  // period of frame
	//int64_t frame_bytes;
	int64_t bytes_frame;  // bytes/frame (including header)
	int64_t pos,wpos;
    vdif_header_t vhead;

    vd->headlen=8;  // default header length (in 32-bit word)

	vd->flag_vdif_anal=0;
	vd->pos_1st_header=0;

	vd->flag_open=0;

	// file open
	if( (vd->fin  = fopen( vd->fname, "rb" )) == NULL ){
         printf("get_real_vdif_info: File (%s) Open Error! \n",vd->fname);
         return -99;
    }
	vd->flag_open=1;
	vd->pos=0;
	// file size get
	vd->f_size=get_file_size(vd->fname);

	if(vd->skip_bytes > 0) {  // skip 1st skip_byte
	   printf(" get_real_vdif_info: Skipping 1st %lld bytes .....",(long long)vd->skip_bytes);
	   fseek(vd->fin,vd->skip_bytes,SEEK_SET);
	   printf("\n");
	}

    // check 1st portion is a possible VDIF header or not 
	if(check_vdif_header(vd->fin) < 0){
	    printf(" get_real_vdif_info: 1st portion is not a VDIF header.\n");
	    printf(" get_real_vdif_info: You want to search header assuming VDIF Kashima format? ([y]/n)");
        read1line(cbuf);
	    if(cbuf[0]=='n' || cbuf[0]=='N'){ // yes
           i=fclose(vd->fin);
	       vd->flag_open=0;
           return -1;
		} else {
		   vd->flag_compulsory=1;
		}
	}

	if(vd->flag_compulsory > 0){
	    if(find_Kas_sync(vd->fin, &vd->pos, 1)<0){
           printf(" get_real_vdif_info: Data File does not have VDIF Kashima header!\n");
           i=fclose(vd->fin);
	       vd->flag_open=0;
           return -1;
		}
		printf(" get_real_vdif_info: Data file has VDIF Kashima header.\n");
		if(vd->pos > 0){
		   printf(" get_real_vdif_info: But 1st header is not a correct header!!\n");
		   printf(" get_real_vdif_info: 1st correct header is located at %lld byte position\n",(long long)vd->pos);
		}
	}


	// header read

	wpos=ftell(vd->fin);  // save current pointer position
	vd->pos_1st_header=wpos;  // 1st correct header position

	iread = fread( vd->head, sizeof(u_int32_t), vd->headlen, vd->fin );  // header read
    if (iread ==0 ) { // EOF
        i=fclose(vd->fin);
	    vd->flag_open=0;
        return -1;
    }
	//vd->pos+=iread*sizeof(u_int32_t);
	parse_vdif_header(vd->head, &vd->vhead);   // 1st header

	// keep 1st headder info
	vhead=vd->vhead;

	// set read pointer at the first of this frame
    fseek(vd->fin,wpos,SEEK_SET); // move pointer to the 1st of file
    if( ferror(vd->fin) ){
        i=fclose(vd->fin);
	    vd->flag_open=0;
        return -1;
    }
	//vd->pos=0;

	if(vd->vhead.legacy==1) vd->headlen=4;  // in 32-bit word
	ref_thread_id=vd->vhead.thread_id;

	vd->adbit=vd->vhead.bpsampl+1;
	vd->numch=(int)pow(2.0, vd->vhead.log2_chs);
	if(vd->vhead.edv==1){  // Kashima EDV
		if(vd->vhead.uflag==0){
			vd->fsample=vd->vhead.srate*1.0e3;
		} else {
			vd->fsample=vd->vhead.srate*1.0e6;
		}
	} else {  // get information by reading 1 sec data later
		vd->fsample=0.0;
	}
	//vd->dtimest=vd->vhead.t.dtime;   // time in day at start


	// calc total number of frames from file size and bytes/frame data
	vd->tnframe=(double)vd->f_size/(double)(vd->vhead.frame_len64*8);

    vd->frame_bytes=(vd->vhead.frame_len64*2-vd->headlen)*4;    // supposed frame bytes excepting header
    //vd->frame_bytes=vd->vhead.frame_len64*8;    // supposed frame bytes including header  2018-12-11
	sec_old=vd->vhead.sec_fm_ref;


	// set Extended format
	edv=vhead.edv;
	if(vd->flag_compulsory > 0) {
	   vd->headlen=8;
	   edv=1; // compulsory set Kashima mode
	   if(vd->flag_compulsory >1) bytes_frame=vd->flag_compulsory;
	}

//    if(mode==0){  //   automatic mode
//        if(vd->vhead.edv==1){  // Kashima EDV
//            if(vd->fsample != 0.0) {  // sampling frequency was already set from header info
//                pframe=(double)vd->frame_bytes*8.0/((double)(nch*adbit)*vd->fsample);  // period of frame
//                vd->nframsec=(int)(1.0/pframe);
//            }
//        }
//    }

	nfram_per_sec=0; 
	nfram_per_sec_flag=0;
	frame_c=0;

	nthread=1;
	thread_flag=1;
	//thread_id_old=-1;

    while(1){
	    while(1){
	         //wpos=ftell(vd->fin);  // save current pointer position  2018-12-11 moved lower

             //iread = fread( ibuf, sizeof( unsigned int ), headlen, fin );  // 1 frame read
			 iread = fread( vd->head, sizeof( unsigned int ), vd->headlen, vd->fin );  // 1 frame headder             
			 if (iread ==0 ) { // EOF
	            printf("\n");
                goto exithere; 
             }
	         wpos=ftell(vd->fin);  // save current pointer position  2018-12-11  moved from above
             parse_vdif_header(vd->head, &vd->vhead);

			 // check header shync block in case of compulsory set Kashima mode
	         //if(runp->flag_compulsory > 0) {
	         if(edv==1) {
		        if(vd->vhead.w2 != syncpat){
		           if(find_Kas_sync(vd->fin, &pos, 0)<0){  // header search with mode=0
		           //if(find_Kas_sync(vd->fin, &pos, 1)<0){  // header search with mode=1
                       printf(" get_real_vdif_info: VDIF Kashima header can not be found in %d words!\n",HEAD_SEARCH_RANGE);
					   printf(" get_real_vdif_info:  search header further? ([y]/n) ");
                       read1line(cbuf);
		               if(cbuf[0]=='n' || cbuf[0]=='N'){ // yes
		                   if(find_Kas_sync(vd->fin, &pos, 1)<0){  // header search with mode=1
                               printf(" get_real_vdif_info: VDIF Kashima header can not be found after all\n");
		                       iok=-1;
		                       goto exithere;
						   }
					   } else {
		                   iok=-1;
		                   goto exithere;
					   }
		           }
		           if(pos > 0){
					   //printf(" correct VDIF Kashima header found at %lld byte position\n",(long long)pos);
					   wpos=pos;
		           }
			    } else {
			       goto exit2;
			    }
		     } else {
		        goto exit2;
		     }

	   }

exit2:

       frame_c++;  // total frame number counter


	   if(frame_c > 1 ){
		   if(thread_flag == 1){
			   if(vd->vhead.thread_id != ref_thread_id) {
				   nthread++;
			   } else {
				   thread_flag=0;

                   /* already # of thread get */

                   if(mode==0 && vd->fsample !=0.0) goto exithere; // not have to read any more
			   }
		   }
	   }

	   if(vd->vhead.sec_fm_ref!=sec_old) {  // new second start
		   if(nfram_per_sec_flag==1) nfram_per_sec_flag=0;
		   if(nfram_per_sec > 0) goto exithere;
	   }

	   // #frames counts
	   if(vd->vhead.frame==0){
		   if(vd->vhead.thread_id==ref_thread_id) {
			 if(nfram_per_sec_flag==0 && nfram_per_sec==0) nfram_per_sec_flag=1;
		   }
	   }
	   if(nfram_per_sec_flag==1 && vd->vhead.thread_id==ref_thread_id) nfram_per_sec++;

       //fseek(vd->fin,vd->frame_bytes,SEEK_CUR); // pointer set to next frame
       fseek(vd->fin,wpos+vd->frame_bytes,SEEK_SET); // move pointer to the next frame  2018-02-03
       if( ferror(vd->fin) ){
           i=fclose(vd->fin);
	       vd->flag_open=0;
           return -1;
       }
	   //vd->pos+=vd->frame_bytes;
	   sec_old=vd->vhead.sec_fm_ref;
	   //thread_id_old=vhead.thread_id;
	}

exithere:

    i=fclose(vd->fin);
	vd->flag_open=0;

	vd->flag_vdif_anal=1;
	//vd->pos=0;

    if(mode==0 && vd->fsample !=0.0) {  //  sampling info already got from header
        pframe=(double)vd->frame_bytes*8.0/((double)(vd->adbit*vd->numch)*vd->fsample);  // period of frame
        vd->nframsec=(int)(1.0/pframe+0.5);
        vd->fbps=(double)vd->nframsec*(double)vd->frame_bytes*8.0;  // data rate per thread
    } else {
	    vd->nframsec=nfram_per_sec;  // frame/sec per thread
	    vd->fbps=(double)nfram_per_sec*(double)vd->frame_bytes*8.0;  // data rate per thread
        vd->fsample=vd->fbps/(double)(vd->adbit*vd->numch);
        pframe=(double)vd->frame_bytes*8.0/((double)(vd->adbit*vd->numch)*vd->fsample);  // period of frame
    }
	vd->tsample=1.0/vd->fsample;
    vd->dlength=pframe*vd->tnframe;

	vd->nthread=nthread;
	vd->bytes1sec=vd->vhead.frame_len64*8*vd->nthread*vd->nframsec;  // bytes in 1 sec (including headder and all threads)

	// retreieve 1st header info
	vd->vhead=vhead;

    return 0;

}


int vdif_t_ini(vdif_file_t *vd, int mode)
{
/*
     vdif parameter initialize by reading vdif data
     2014-05-07
     2014-06-12

	 return  0: OK
	        -1: error occurred
*/

    if(get_real_vdif_info(vd, mode)<0) return -1;
	vd->dtimest=vd->vhead.t.dtime;   // time in day at start
	vd->pos=0;

	return 0;
}

int fwdNsecVDIF(int nsec, vdif_file_t *vd)
{
/* 
       forward fast N sec
	   if current position is not on the second,
	   1sec forward means forward at the begining of the next second
	   (so less than 1 sec forward)
       This function does not care about Thread#.

       2018-12-12  revised
*/
	int64_t pos,posk;

    if(nsec < 1) return 0; // nothing to do

    //tword=K5->twordread;  // save current word read counter data 
    //tbytes=K5->tbytesread;  // save current byte read counter data 
    pos=vd->pos;  // save current file pointer


	// compute set position
    if(vd->vhead.frame != 0){     // 2018-12-12
        posk=pos+(vd->nframsec-vd->vhead.frame)*(vd->frame_bytes+32);  // to set on the second
        posk+=(nsec-1)*vd->bytes1sec;
    } else {
    	posk=((int64_t)(pos/vd->bytes1sec)+nsec)*vd->bytes1sec;
    }
	if(posk < vd->f_size){  // 
		fseek(vd->fin,posk,SEEK_SET); // move pointer by K5->bytesec
        if( ferror( vd->fin ) ){
            return -1;
        }
		vd->pos=posk;
    } else {  // nothing to do
        return -1;
    }
       
    return 0;
}


int fwdNsecVDIF_multi(int nsec, vdif_file_t *vd)
{
/* 
       forward fast N sec for multithread data
	   if current position is not on the second,
	   1sec forward means forward at the begining of the next second
	   (so less than 1 sec forward)
       This function cares about Thread#.

       2018-07-23

*/
	int64_t pos,posk;
    int i;
    int isec, csec, csec_old;
    //timeval_vdif_t tc;

    if(nsec < 1) return 0; // nothing to do

    // keep current time label
    //tc=vd->vhead.t;
    isec=vd->vhead.t.sec_fm_ref;

    csec_old=-1;  // for debug
    //pos=vd->pos
    // read headder 
loop:

    if(vdif_header_read_mode1(vd) <0) return -1;   // error return
    if(vd->vhead.thread_id != vd->pup_thread) goto next;
    // check frame #
    if(vd->vhead.frame != 0) goto next;
    csec=vd->vhead.t.sec_fm_ref;
//    if(csec != csec_old) printf(" fwdNsecVDIF_multi: Debug csec %d  isec %d\n",csec,isec);   // debug
    csec_old=csec;                                                          // debug
    if(csec-isec >= nsec) goto exitloop;
next:
    // forward 1 frame + header
    vd->pos+=(vd->frame_bytes+vd->headlen*4);
    fseek(vd->fin,vd->pos,SEEK_SET); // move pointer to the next frame 
    goto loop;

exitloop:
    //fgetpos(vd->fin, &(fpos_t *)vd->pos);
    vd->pos = ftell(vd->fin);

    return 0;
}




int vdif_header_read_mode1(vdif_file_t *vd)
{
/* 
	header read and analyze, then pointer back to the original position
    Initial file pointer should be the begining of header (frame)
*/
	int i;

    //i=fgetpos(vd->fin, &vd->pos);  // 2019-11-01
    i=fgetpos(vd->fin, (fpos_t *)&vd->pos);  // 2019-11-01 (to avoid warning at linux)

	if(fread( vd->head, sizeof(u_int32_t), vd->headlen, vd->fin )==0){  // header read
        i=fclose(vd->fin);
        vd->flag_open=0;
        return -1;
    }
	//vd->pos+=iread*sizeof(u_int32_t);
	parse_vdif_header(vd->head, &vd->vhead);   // 1st header

	// set read pointer to the original point
	fseek(vd->fin,vd->pos,SEEK_SET); // move pointer to the original
    if( ferror(vd->fin) ){
        i=fclose(vd->fin);
        vd->flag_open=0;
        return -1;
    }
	return 0;
}

int vdif_header_read_mode2(vdif_file_t *vd)
{
/* 
	header read and analyze. Pointer doesn't back to the original position
    Initial file pointer should be the begining of header (frame)
*/
	int i;
	if(fread( vd->head, sizeof(u_int32_t), vd->headlen, vd->fin )==0){  // header read
        i=fclose(vd->fin);
        vd->flag_open=0;
        return -1;
    }
	//vd->pos+=iread*sizeof(u_int32_t);
	parse_vdif_header(vd->head, &vd->vhead);   // 1st header

	//vd->pos+=vd->headlen*4;    // use below
    //fgetpos(vd->fin, &vd->pos);  // 2019-11-01
    i = fgetpos(vd->fin, (fpos_t *)&vd->pos);  // 2019-11-01 (to avoid warning at linux)
    return 0;
}

int vdif_header_read_check(vdif_file_t *vd, int mode, int64_t *sa_pos)
{
/* 
	read 32 byte then judge whether it is a correct header or not.
	if mode is set to search Kashima header, header search according to
	mode forward.

	if header is a correct one, pointer set to the beginning of header block


	[in]
	     vd->flag_compulsory -- 0: normal VDIF
		                       +VE: Compulsory set Kashima VDIF format
         mode --- sync search mode in case of Kashima VDIF format
		          0: limit around given 4byte word range (default)
				  1: search forward (until EOF)
				  2: search back ward (unti Begining of File)

	[return]
	     sa_pos -- byte offset of the current pointer between after call and before call
	        0:  may be VDIF header
		   -1:  not a VDIF header

*/

	int iok;
	int64_t wpos,spos;
    //vdif_header_t vhead;
	iok=-1;
	wpos=ftell(vd->fin);  // save current pointer position
	spos=wpos;
	if(vd->flag_compulsory > 0) {  // Kashima VDIF
		if(find_Kas_sync(vd->fin, &vd->pos, mode) < 0) {//   not found Kashima VDIF header
			goto exitreturn;
		}
		spos=vd->pos;
		goto exithere;
	} 
exithere:
	if(fread( vd->head, sizeof(u_int32_t), vd->headlen, vd->fin )==0){  // header read
       // EOF
       goto exitreturn; 
    }
	parse_vdif_header(vd->head, &vd->vhead);
	if(vd->vhead.frame_len64 <= 0) goto exitreturn;
	if(vd->vhead.t.year <= 0 || vd->vhead.t.year > 2050) goto exitreturn;  // not VDIF header
	iok=0;
exitreturn:
    fseek(vd->fin,spos,SEEK_SET);
	vd->pos=spos;
	*sa_pos=spos-wpos;
	return iok;
}


int vdif_read_unit(vdif_file_t *vd, int r_thread_id)  // 2014-05-08  2018-02-03
{
/*   
    2018-02-03  revised to support # of CH > 16  data max is 64 channels

    Note!!  2018-02-11 must be revised to handle the channel of I, Q data correctly

    This function read unit sample data for all channles continuously from vdif file

    At 1st call,  file should be opened, pointer should be at the first point
	or begining of header, and flags and variables under "vd" should be set as follows.
         
       flag_read_header = 1   --  file read pointer is set at the begining of VDIF header
       flag_read_data   = 1   --  should read data
       flag_idata_over  = 0   
       usampl_c         = 0

    After second call, don't touch above two flags.

    Operation:
    At 1st call, header is read, and then data field is read until the next header
    by "imax" times by "nword" x 32bit-word, as follows.


    [Data Reading]

    <------------------------------------- 1 frame data ---------------------------->

     header                              data field                                   header
    I-------I-------------------|-------------------|---   ...... --|----------------I-------I------
                  nword (#1)          nword (#2)                      nword (#imax)
            <--------------- imax * nword (32-bit word unit) ------------>


    Data picked up from nword(*32bits) data by adbit*numch(bits) usampl-times, as follows. 

    [Data Encoding]

    <---------------------------------- nword (#1) ------------------------------------------------------>

    |-------------------|-------------------|----  ...... --|-------------------|-------------------|--.....
      adbit*numch(bits)   adbit*numch(bits)                   adbit*numch(bits)   adbit*numch(bits)
         sample #1           sample #2                          sample #usampl      sample #1

    
    If nword data is over before picking up usampl-times data,  necxt nword data are used for picking up data.



    [in]
	    r_thread_id  --- thread id# to read


    [return]
             -99:  EOF
	         -1 :  other errors
			  0 :  OK
         flag_unit_read_ok  -- 1: Unit data 

*/

    int pat, n, ilast,k,m;
    int pattbl[9]={1,1,3,3,15,15,15,15,255};
    u_int32_t wd1;
	int mode;
    int64_t sa_pos;
	int64_t wpos;  // debug
    //int *VDpos;
    //int wdata[32];
    int wdata[64];   // allow max 64 channel  2018-02-03
    int tmpdat[32];  // 2018-02-03
    int nloop;    // 2018-02-03
    int uposadd;  // 2018-02-03
    int ilast2;   // 2018-02-03
    int bitstream;  // # of bits for complete one sample   = adbit*CHs for real data, 2*adbit*CHs for complex data

    int flag_complex;  // 2018-02-11
    int kmax;          // 2018-02-11

	//initialize
    ilast=32/vd->adbit;
    pat=pattbl[vd->adbit];

    flag_complex=vd->vhead.d_type; // retrieve flag_complex for after second call


    bitstream=vd->adbit*vd->numch;     // 2018-03-05
    if(flag_complex==1) bitstream*=2;  // 2018-03-05

    nloop=1;      // 2018-02-03
    //if(vd->adbit*vd->numch > 32) nloop=2;   // 2018-02-03
    //if(vd->adbit*vd->numch > 64) nloop=3;   // 2018-02-03
    //if(vd->adbit*vd->numch > 96) nloop=4;   // 2018-02-03
    //nloop=(vd->adbit*vd->numch-1)/32 + 1;         // # of 32bit-words for 1 complete sample including all channels 
    nloop=(bitstream-1)/32+1;  //# of 32bit-words for 1 complete sample including all channels   2018-03-05

    uposadd=nloop;                          // 2018-02-03
    ilast2=32*nloop/vd->adbit;              // last position in a complete sample

    //bpat=(unsigned char)pat;

	vd->flag_unit_read_ok=0;
	
	if(vd->flag_idata_over > 0) {// copy overeflowed data to idata
        for(n=0; n<vd->flag_idata_over; n++) {
            for(m=0; m<vd->numch; m++){
                vd->ch[m].idata[n]=vd->ch[m].idata_over[n];
            }
        }
		vd->usampl_c=vd->flag_idata_over;
	}

	mode=1;

	// header read
read_header:

	if(vd->flag_read_header==1){ // header read
		if(vdif_header_read_check(vd, mode, &sa_pos)<0){ // header not found
			return -99;
		}
        flag_complex=vd->vhead.d_type; // 2018-02-11
        //flag_complex=1;                // 2018-02-11  for Debug
//        if(flag_complex==1){    // 2018-02-11
//           nloop=(2*vd->adbit*vd->numch-1)/32 + 1;
//           uposadd=nloop; 
//           ilast2=32*nloop/vd->adbit;
//        }


		vd->tot_frame++;   // total frame_counter updated
		if(sa_pos !=0){
			printf(" vdif_read_unit: warning!!  Header found at %lld bytes shifted position\n",(long long)sa_pos);
			printf("                               for %lld-th frame!!\n",(long long)vd->tot_frame);
		}
		if(vd->vhead.thread_id != r_thread_id){ //  move to next thread
			fseek(vd->fin,vd->vhead.frame_len64*8,SEEK_CUR);
            if( ferror(vd->fin) ){
				return -1;
			}
			goto read_header;
		}
		vd->pos+=vd->headlen*4;   // pointer update to read data portion
        fseek(vd->fin,vd->pos,SEEK_SET);  
        if( ferror(vd->fin) ){
			return -1;
		}

        //parse_vdif_header(vd->head, &vd->vhead);
		//vd->dtime=vd->vhead.t.dtime;
		vd->flag_read_header=0;
	}

	// data read
read_data:

	if(vd->flag_read_data==1){ // data read
		if(vd->imax_c >= vd->imax){    //  go to header read
			vd->imax_c=0;
			vd->flag_read_header=1;
			goto read_header;
		}

	    wpos=ftell(vd->fin);  // debug
        vd->wordread = fread(vd->ubuf, sizeof(u_int32_t), vd->nword, vd->fin);
		if (vd->wordread ==0 ) { // EOF
			return -99;
		}
	    wpos=ftell(vd->fin);  // debug
        vd->pos+=vd->wordread*sizeof(u_int32_t);  // current file pointer position (bytes)
		vd->imax_c++;
		vd->upos=0;
		vd->flag_read_data=0;
	}

	// unit data pick up and decode to sampling data
decode_data:

	// 32bit-word data decode

	//wd1=vd->ubuf[vd->upos];
    //decode_data(wdata, wd1, ilast, pat, vd->adbit);   // use of inline assembler

    // look again carefully!!

    for(k=0; k<nloop; k++){
	   wd1=vd->ubuf[vd->upos];
       decode_data(tmpdat, wd1, ilast, pat, vd->adbit);   // use of inline assembler     
       //for(m=0; m<ilast; m++) wdata[k*ilast+m]=tmpdat[m];   // 2018-02-03
       memcpy(wdata+k*ilast,tmpdat,ilast*sizeof(int));      // 2018-02-03
       //memcpy(wdata+k*ilast,tmpdat,32);      // 2018-02-03
       vd->upos++;
    }
    vd->upos--;                                        // 2018-02-03

//    printf("wdata: ");
//    for(k=0; k<nloop*ilast; k++) printf(" %d",wdata[k]);  // debug
//    printf("\n");
    // put data into each channle data   (should consider in case of I, Q data) 

    //for(k=0; k<ilast/vd->numch; k++){ //  
    kmax=ilast2/vd->numch;   // 2018-02-11
    if(flag_complex==1) kmax/=2;  // 2018-02-11  !! not good   !!
    //for(k=0; k<ilast2/vd->numch; k++){ //     2018-02-03      // when I, Q data  k< (ilast2/vd->numch)/2  ?
    for(k=0; k<kmax; k++){ //     2018-02-03      // when I, Q data  k< (ilast2/vd->numch)/2  ?
        for(m=0; m<vd->numch; m++){
           if(flag_complex==0){
               vd->ch[m].idata[vd->usampl_c]=wdata[k*vd->numch+m];
           } else {
               vd->ch[m].idata[vd->usampl_c]=wdata[2*k*vd->numch+2*m];     // when I, Q data  I =wdata[2*k*vd->numch+2*m]  is this OK? 
               vd->ch[m].idata[vd->usampl_c+1]=wdata[2*k*vd->numch+2*m+1]; //                 Q =wdata[2*k*vd->numch+2*m +1]  ?
           }                                                                                                         
       }
       vd->usampl_c++;
       if(flag_complex==1) vd->usampl_c++;       // 2018-02-11
    }

//    if((vd->usampl_c % 2) !=0 ){     // debug
//        printf(" Debug %d vd->usampl_c %d\n",flag_complex, vd->usampl_c);
//    }

	vd->npos=vd->usampl_c;
	//if(vd->usampl_c >= vd->usampl){  // data ready    !! must consider in case that usampl is not match with the boundary
    if(vd->usampl_c >= vd->usampl){  // data ready    !! must consider in case that usampl is not match with the boundary
		                                           // of 32bit word
		if(vd->usampl_c > vd->usampl){
			vd->npos=vd->usampl;
			vd->flag_idata_over=vd->usampl_c-vd->usampl;
            for(n=0; n<vd->flag_idata_over; n++){
                for(m=0; m<vd->numch; m++){
                   vd->ch[m].idata_over[n]=vd->ch[m].idata[vd->usampl+n];
                }
            }
		} else {
			vd->flag_idata_over=0;
		}
		vd->usampl_c=0;
		vd->flag_unit_read_ok=1;
		vd->n_usampl++;
		vd->dtime=vd->dtimest+(double)(vd->n_usampl-1)*vd->unitp/86400.0;  // set time at the begining of unit
		//if(fmod(fmod(vd->dtime,1.0)*86400.0,1.0)<1.0e-6 || fmod(fmod(vd->dtime,1.0)*86400.0,1.0)> 1.0-1.0e-6){//debug
		//	printf(" Debug %f\n",fmod(vd->dtime,1.0)*86400.0);
		//	printf(" Debug OK\n");
		//}
        for(m=0; m<vd->numch; m++){
            vd->ch[m].usampl=vd->npos;
        }
		goto exit2return;
	}
	vd->upos++;
	if(vd->upos >= vd->wordread) {
	    vd->flag_read_data=1;
		if(vd->imax_c >= vd->imax) {
	        vd->flag_read_header=1;
			vd->imax_c=0;
			goto read_header;
		}

		goto read_data;
	}

	goto decode_data;

exit2return:
	return 0;
}



int vdif_read_unit_old(vdif_file_t *vd, int r_thread_id)  // 2014-05-08  2018-02-03  save 2018-02-11 befor revising
{
/*   
    2018-02-03  revised to support # of CH > 16  data max is 64 channels

    Note!!  2018-02-11 must be revised to handle the channel of I, Q data corrdctly

    This function read unit sample data for all channles continuously from vdif file

    At 1st call,  file should be opened, pointer should be at the first point
	or begining of header, and flags and variables under "vd" should be set as follows.
         
       flag_read_header = 1   --  file readi pointer is set at the begining of VDIF header
       flag_read_data   = 1   --  should read data
       flag_idata_over  = 0   
       usampl_c         = 0

    After second call, don't touch above two flags.

    Operation:
    At 1st call, header is read, and then data field is read until the next header
    by "imax" times by "nword" x 32bit-word, as follows.


    [Data Reading]

    <------------------------------------- 1 frame data ---------------------------->

     header                              data field                                   header
    I-------I-------------------|-------------------|---   ...... --|----------------I-------I------
                  nword (#1)          nword (#2)                      nword (#imax)
            <--------------- imax * nword (32-bit word unit) ------------>


    Data picked up from nword(*32bits) data by adbit*numch(bits) usampl-times, as follows. 

    [Data Encoding]

    <---------------------------------- nword (#1) ------------------------------------------------------>

    |-------------------|-------------------|----  ...... --|-------------------|-------------------|--.....
      adbit*numch(bits)   adbit*numch(bits)                   adbit*numch(bits)   adbit*numch(bits)
         sample #1           sample #2                          sample #usampl      sample #1

    
    If nword data is over before picking up usampl-times data,  necxt nword data are used for picking up data.



    [in]
	    r_thread_id  --- thread id# to read


    [return]
             -99:  EOF
	         -1 :  other errors
			  0 :  OK
         flag_unit_read_ok  -- 1: Unit data 

*/

    int pat, n, ilast,k,m;
    int pattbl[9]={1,1,3,3,15,15,15,15,255};
    u_int32_t wd1;
	int mode;
    int64_t sa_pos;
	int64_t wpos;  // debug
    //int *VDpos;
    //int wdata[32];
    int wdata[64];   // allow max 64 channel  2018-02-03
    int tmpdat[32];  // 2018-02-03
    int nloop;    // 2018-02-03
    int uposadd;  // 2018-02-03
    int ilast2;   // 2018-02-03

	//initialize
    ilast=32/vd->adbit;
    pat=pattbl[vd->adbit];

    nloop=1;      // 2018-02-03
    //if(vd->adbit*vd->numch > 32) nloop=2;   // 2018-02-03
    //if(vd->adbit*vd->numch > 64) nloop=3;   // 2018-02-03
    //if(vd->adbit*vd->numch > 96) nloop=4;   // 2018-02-03
    nloop=(vd->adbit*vd->numch-1)/32 + 1;         // 2018-02-03
    uposadd=nloop;                          // 2018-02-03
    ilast2=32*nloop/vd->adbit;              // 2018-02-03

    //bpat=(unsigned char)pat;

	vd->flag_unit_read_ok=0;
	
	if(vd->flag_idata_over > 0) {// copy overeflowed data to idata
        for(n=0; n<vd->flag_idata_over; n++) {
            for(m=0; m<vd->numch; m++){
                vd->ch[m].idata[n]=vd->ch[m].idata_over[n];
            }
        }
		vd->usampl_c=vd->flag_idata_over;
	}

	mode=1;

	// header read
read_header:

	if(vd->flag_read_header==1){ // header read
		if(vdif_header_read_check(vd, mode, &sa_pos)<0){ // header not found
			return -99;
		}
		vd->tot_frame++;   // total frame_counter updated
		if(sa_pos !=0){
			printf(" vdif_read_unit: warning!!  Header found at %lld bytes shifted position\n",(long long)sa_pos);
			printf("                               for %lld-th frame!!\n",(long long)vd->tot_frame);
		}
		if(vd->vhead.thread_id != r_thread_id){ //  move to next thread
			fseek(vd->fin,vd->vhead.frame_len64*8,SEEK_CUR);
            if( ferror(vd->fin) ){
				return -1;
			}
			goto read_header;
		}
		vd->pos+=vd->headlen*4;   // pointer update to read data portion
        fseek(vd->fin,vd->pos,SEEK_SET);  
        if( ferror(vd->fin) ){
			return -1;
		}

        //parse_vdif_header(vd->head, &vd->vhead);
		//vd->dtime=vd->vhead.t.dtime;
		vd->flag_read_header=0;
	}

	// data read
read_data:

	if(vd->flag_read_data==1){ // data read
		if(vd->imax_c >= vd->imax){    //  go to header read
			vd->imax_c=0;
			vd->flag_read_header=1;
			goto read_header;
		}

	    wpos=ftell(vd->fin);  // debug
        vd->wordread = fread(vd->ubuf, sizeof(u_int32_t), vd->nword, vd->fin);
		if (vd->wordread ==0 ) { // EOF
			return -99;
		}
	    wpos=ftell(vd->fin);  // debug
        vd->pos+=vd->wordread*sizeof(u_int32_t);  // current file pointer position (bytes)
		vd->imax_c++;
		vd->upos=0;
		vd->flag_read_data=0;
	}

	// unit data pick up and decode to sampling data
decode_data:

	// 32bit-word data decode

	//wd1=vd->ubuf[vd->upos];
    //decode_data(wdata, wd1, ilast, pat, vd->adbit);   // use of inline assembler

    for(k=0; k<nloop; k++){
	   wd1=vd->ubuf[vd->upos];
       decode_data(tmpdat, wd1, ilast, pat, vd->adbit);   // use of inline assembler
       //for(m=0; m<ilast; m++) wdata[k*ilast+m]=tmpdat[m];   // 2018-02-03
       memcpy(wdata+k*ilast,tmpdat,ilast*sizeof(int));      // 2018-02-03
       //memcpy(wdata+k*ilast,tmpdat,32);      // 2018-02-03
       vd->upos++;
    }
    vd->upos--;                                        // 2018-02-03

    // put data into each channle data   (sould consider in case of I, Q data) 
    //for(k=0; k<ilast/vd->numch; k++){ //  
    for(k=0; k<ilast2/vd->numch; k++){ //     2018-02-03
       for(m=0; m<vd->numch; m++){
            vd->ch[m].idata[vd->usampl_c]=wdata[k*vd->numch+m];
       }
       vd->usampl_c++;
    }


	vd->npos=vd->usampl_c;
	//if(vd->usampl_c >= vd->usampl){  // data ready    !! must consider in case that usampl is not match with the boundary
    if(vd->usampl_c >= vd->usampl){  // data ready    !! must consider in case that usampl is not match with the boundary
		                                           // of 32bit word
		if(vd->usampl_c > vd->usampl){
			vd->npos=vd->usampl;
			vd->flag_idata_over=vd->usampl_c-vd->usampl;
            for(n=0; n<vd->flag_idata_over; n++){
                for(m=0; m<vd->numch; m++){
                   vd->ch[m].idata_over[n]=vd->ch[m].idata[vd->usampl+n];
                }
            }
		} else {
			vd->flag_idata_over=0;
		}
		vd->usampl_c=0;
		vd->flag_unit_read_ok=1;
		vd->n_usampl++;
		vd->dtime=vd->dtimest+(double)(vd->n_usampl-1)*vd->unitp/86400.0;  // set time at the begining of unit
		//if(fmod(fmod(vd->dtime,1.0)*86400.0,1.0)<1.0e-6 || fmod(fmod(vd->dtime,1.0)*86400.0,1.0)> 1.0-1.0e-6){//debug
		//	printf(" Debug %f\n",fmod(vd->dtime,1.0)*86400.0);
		//	printf(" Debug OK\n");
		//}
        for(m=0; m<vd->numch; m++){
            vd->ch[m].usampl=vd->npos;
        }
		goto exit2return;
	}
	vd->upos++;
	if(vd->upos >= vd->wordread) {
	    vd->flag_read_data=1;
		if(vd->imax_c >= vd->imax) {
	        vd->flag_read_header=1;
			vd->imax_c=0;
			goto read_header;
		}

		goto read_data;
	}

	goto decode_data;

exit2return:
	return 0;
}






// add 2018-02-03


int vdif_thread_anal(vdif_file_t *vd, int *numthread, int *table_thread, int *table_frames)
{
/*
       Ver.2018-02-03
           2019-12-11 add function name in printf statements

       analyze # of threads in VDIF file

   [out]
       numthread   --  # of threads
       table_thread[ ] -- thread# table
       table_frames[ ] -- # of frames during thread ID doesn't change  (# of frames per second for the case of numthread=1)

   [return value]
           +VE:   # of threads
	        -1: not a VDIF format
	       -99: file open error or file read error
*/

    //u_int32_t head[8];
    unsigned int syncpat=VDIF_KASHIMA_SYNC;    // sync pattern for VDIF Kashima format
	char cbuf[10]; // for operator interface (if necessary)
	int i,iread;
	//int headlen; 
	int edv;
	int iok;
    int thread_num_count;
    int s_table[50];  // work thread# table
    int s_frame_count[50];  // frame counter by thread
    int s_1st_frame_num[50]; // 1 st frame# by thread
    int s_frame_num[50]; // current (just read) frame# by thread
    int s_flag_1st[50]; // flag to show 1st frame by thread
    int max_thread=50;   // max # of threads
    int s_index, s_index_old;
	int frame_c;
	int sec_old;
	int ref_thread_id;
	int nfram_per_sec;  //  true # of frames per sec
	int nfram_per_sec_flag;  //  true # of frames per sec count OK flag
	int nthread; // true # of threads
	int thread_id_1st;
    int thread_id_old;
    int frame_num_1st;
    int flag_1st;
	//int thread_id_old;
	int thread_flag;
    double pframe;  // period of frame
	//int64_t frame_bytes;
	int64_t bytes_frame;  // bytes/frame (including header)
	int64_t pos,wpos;
    vdif_header_t vhead;

    vd->headlen=8;  // default header length (in 32-bit word)

	vd->flag_vdif_anal=0;
	vd->pos_1st_header=0;

	vd->flag_open=0;

    for(i=0; i<max_thread; i++){
        s_table[i]=-1;
        s_frame_count[i]=0;
        s_1st_frame_num[i]=-1;
        s_frame_num[i]=-1;
        s_flag_1st[i]=1;
    }
    s_index_old=-1;

    thread_num_count=0;
    thread_id_1st=-1;
    thread_id_old=-1;

	// file open
	if( (vd->fin  = fopen( vd->fname, "rb" )) == NULL ){
        printf("vdif_thread_anal: File (%s) Open Error! \n",vd->fname);
        return -99;
    }
	vd->flag_open=1;
	vd->pos=0;
	// file size get
	vd->f_size=get_file_size(vd->fname);

	if(vd->skip_bytes > 0) {  // skip 1st skip_byte
	   printf(" vdif_thread_anal: Skipping 1st %lld bytes .....",(long long)vd->skip_bytes);
	   fseek(vd->fin,vd->skip_bytes,SEEK_SET);
	   printf("\n");
	}

    // check 1st portion is a possible VDIF header or not 
	if(check_vdif_header(vd->fin) < 0){
	    printf(" vdif_thread_anal: 1st portion is not a VDIF header.\n");
	    printf(" vdif_thread_anal: You want to search header assuming VDIF Kashima format? ([y]/n)");
        read1line(cbuf);
	    if(cbuf[0]=='n' || cbuf[0]=='N'){ // yes
           i=fclose(vd->fin);
	       vd->flag_open=0;
           return -1;
		} else {
		   vd->flag_compulsory=1;
		}
	}
	if(vd->flag_compulsory > 0){
	    if(find_Kas_sync(vd->fin, &vd->pos, 1)<0){
           printf(" vdif_thread_anal: Data File does not have VDIF Kashima header!\n");
           i=fclose(vd->fin);
	       vd->flag_open=0;
           return -1;
		}
		printf(" vdif_thread_anal: Data file has VDIF Kashima header.\n");
		if(vd->pos > 0){
		   printf(" vdif_thread_anal: But 1st header is not a correct header!!\n");
		   printf(" vdif_thread_anal: 1st correct header is located at %lld byte position\n",(long long)vd->pos);
		}
	}

    flag_1st=1;

    frame_c=0;
    // frame read loop

    while(1){
	    wpos=ftell(vd->fin);  // save current pointer position

        iread = fread( vd->head, sizeof( unsigned int ), vd->headlen, vd->fin );  // 1 frame headder             
        if (iread ==0 ) { // EOF
	       printf("\n");
           goto exithere; 
        }
        parse_vdif_header(vd->head, &vd->vhead);
        vd->frame_bytes=vd->vhead.frame_len64*8;    // supposed frame bytes excepting header

//        if(flag_1st==0){
//            if(vd->vhead.thread_id == thread_id_1st && thread_num_count > 1) goto exithere; 
//        }

        // register new thread number
        if(vd->vhead.thread_id != thread_id_old){
            //if(thread_num_count == 0){
            //    s_table[thread_num_count]=vd->vhead.thread_id;
            //    thread_num_count++;
            //}
            s_index=-1;
            for(i=0; i<thread_num_count; i++){
                if(vd->vhead.thread_id == s_table[i]) {
                   s_index=i;
                   break;
                }
            }
            if(s_index < 0) { // register new thread ID
               s_table[thread_num_count]=vd->vhead.thread_id;
               s_index=thread_num_count;
               thread_num_count++;
            } else { // already appeared before
                goto exithere;
            }
        }

        // at least 1 sec read in case of a single thread
        if(thread_num_count == 1){
            if(s_frame_count[s_index] > 0 && vd->vhead.frame ==  s_1st_frame_num[s_index]) goto exithere;  //  1sec passed
        }


        if(s_flag_1st[s_index]==1){ // this is the 1st frame of the thread
            s_1st_frame_num[s_index]=vd->vhead.frame;
            s_frame_count[s_index]=0;
            s_flag_1st[s_index]=0;
        }

        s_frame_num[s_index]=vd->vhead.frame;
        s_frame_count[s_index]++;

        //frame_c++;  // total frame number counter
        
        // forward one frame

        //fseek(vd->fin,vd->frame_bytes,SEEK_CUR); // pointer set to next frame
        fseek(vd->fin,wpos+vd->frame_bytes,SEEK_SET); // move pointer to the benext frame  2018-02-03
        if( ferror(vd->fin) ){
            i=fclose(vd->fin);
	        vd->flag_open=0;
            return -99;
        }
        thread_id_old=vd->vhead.thread_id;
        s_index_old=s_index;

    }

exithere:

    i=fclose(vd->fin);
	vd->flag_open=0;

    *numthread=thread_num_count;
    for(i=0; i<thread_num_count; i++){
        table_thread[i]=s_table[i];
        table_frames[i]=s_frame_count[i];
    }

    // debug
//    for(i=0; i<thread_num_count; i++){
//        printf("thread ID = %d\n", s_table[i]);
//        printf("s_frame_count = %d\n", s_frame_count[i]);
//        printf("s_1st_frame_num = %d\n", s_1st_frame_num[i]);
//        printf("s_frame_num = %d\n", s_frame_num[i]);
//    }


    //return 0;
    return thread_num_count;

}



/** add on 2018-06-20 **/


int vdif_read_frame(vdif_file_t *vd, int r_thread_id) 
{
/*   
    2018-06-20  read 1 frame data


    before calling this function, pointer must be the begining of header

    [in]
	    r_thread_id  --- thread id# to read


    [return]
        vd->ubuf  --- 1 frame dat     (size of ubuf should be the size of frame excepting header)
             -99:  EOF
	         -1 :  other errors
			  0 :  OK
         flag_unit_read_ok  -- 1: Unit data 

*/

    int pat, ilast,k,m;
    int pattbl[9]={1,1,3,3,15,15,15,15,255};
    u_int32_t wd1;
	int mode;
    int64_t sa_pos;
	int64_t wpos;  // debug
    //int *VDpos;
    //int wdata[32];
    //int wdata[64];   // allow max 64 channel  2018-02-03
    int tmpdat[32];  // 2018-02-03
    int nloop;    // 2018-02-03
    int uposadd;  // 2018-02-03
    //int ilast2;   // 2018-02-03
    int bitstream;  // # of bits for complete one sample   = adbit*CHs for real data, 2*adbit*CHs for complex data

    int flag_complex;  // 2018-02-11
    int kmax;          // 2018-02-11

	//initialize
    ilast=32/vd->adbit;
    pat=pattbl[vd->adbit];

    flag_complex=vd->vhead.d_type; // retrieve flag_complex for after second call


    bitstream=vd->adbit*vd->numch;     // 2018-03-05
    if(flag_complex==1) bitstream*=2;  // 2018-03-05

    //bpat=(unsigned char)pat;

	vd->flag_frame_read_ok=0;
	
	mode=1;

	// header read
read_header:

	if(vd->flag_read_header==1){ // header read
		if(vdif_header_read_check(vd, mode, &sa_pos)<0){ // header not found
			return -99;
		}
        flag_complex=vd->vhead.d_type; // 2018-02-11

		if(sa_pos !=0){
			printf(" vdif_read_unit: warning!!  Header found at %lld bytes shifted position\n",(long long)sa_pos);
			printf("                               for %lld-th frame!!\n",(long long)vd->tot_frame);
		}
		if(vd->vhead.thread_id != r_thread_id){ //  move to next thread
			fseek(vd->fin,vd->vhead.frame_len64*8,SEEK_CUR);
            if( ferror(vd->fin) ){
				return -1;
			}
		    vd->tot_frame++;   // total frame_counter including other thread #s updated
			goto read_header;
		}
		vd->pos+=vd->headlen*4;   // pointer update to read data portion
        fseek(vd->fin,vd->pos,SEEK_SET);  
        if( ferror(vd->fin) ){
			return -1;
		}

        //parse_vdif_header(vd->head, &vd->vhead);
		//vd->dtime=vd->vhead.t.dtime;
		vd->flag_read_header=0;
        vd->flag_read_data=1;
	}

	// data read
read_data:

	if(vd->flag_read_data==1){ // data read

	    wpos=ftell(vd->fin);  // debug
        vd->wordread = fread(vd->ubuf, sizeof(u_int32_t), vd->nword, vd->fin);
		if (vd->wordread ==0 ) { // EOF
			return -99;
		}
		vd->tot_frame++;   // total frame_counter including other thread #s updated
	    wpos=ftell(vd->fin);  // debug
        vd->pos+=vd->wordread*sizeof(u_int32_t);  // current file pointer position (bytes)
		//vd->imax_c++;
		vd->flag_read_data=0;
		vd->flag_read_header=1;
	}


exit2return:
	return 0;
}


/***  added on 2019-10-23 and later ***/

int fwdNframeVDIF_multi(int nframe, vdif_file_t *vd)
{
    /*
           forward N frames in the same thread# for multithread data


           2019-10-23
           2019-10-31 debug

    */
    int isec, csec_old;
    int count;
    //timeval_vdif_t tc;

    if (nframe < 1) return 0; // nothing to do

    // keep current time label
    //tc=vd->vhead.t;
    isec = vd->vhead.t.sec_fm_ref;

    csec_old = -1;  // for debug
    count = 0;
    //pos=vd->pos
    // read headder 
loop:

    if (vdif_header_read_mode1(vd) < 0) return -1;   // error return
    if (vd->vhead.thread_id != vd->pup_thread) goto next;
    //count++;   deleted on 2019-10-31
    if (count == nframe) goto exitloop;
    // check frame #
    //if (vd->vhead.frame != 0) goto next;
    //csec = vd->vhead.t.sec_fm_ref;
    //if (csec != csec_old) printf("fwdNsecVDIF_multi: Debug csec %d  isec %d\n", csec, isec);   // debug
    //csec_old = csec;                                                          // debug
    //if (csec - isec >= nsec) goto exitloop;
next:
    // forward 1 frame + header
    vd->pos += (vd->frame_bytes + vd->headlen * 4);
    fseek(vd->fin, vd->pos, SEEK_SET); // move pointer to the next frame 
    count++;  // 2019-10-31
    goto loop;

exitloop:

    return 0;
}



int vdif_header_read_mode1_multi(vdif_file_t *vd)
{
    /*
        header read and analyze, then pointer back to the start of frame
        (read given thread# header)
        2019-10-24
    */
    int i;
    //fpos_t pos;
    int64_t hpos;

    //fseek(vd->fin, vd->pos, SEEK_SET); // move pointer to initial position (just in case) 

loop:
    //i = fgetpos(vd->fin, &pos);
    // hpos=(int64_t)pos;  //<== this results in error at linux!!
    // to avoid error on linux, do as follows.

    i = fgetpos(vd->fin, (fpos_t *)&vd->pos);  // 2019-11-01 (to avoid warning at linux)
    hpos = vd->pos;
    if (fread(vd->head, sizeof(u_int32_t), vd->headlen, vd->fin) == 0) {  // header read
        i = fclose(vd->fin);
        vd->flag_open = 0;
        return -1;
    }
    vd->pos +=  vd->headlen * 4;
    parse_vdif_header(vd->head, &vd->vhead);   // 1st header
    if (vd->vhead.thread_id == vd->pup_thread) goto exit_loop;
    // forward to next frame start
    vd->pos += vd->frame_bytes;
    fseek(vd->fin, vd->pos, SEEK_SET); // move pointer to the next frame 
    if (ferror(vd->fin)) {
        i = fclose(vd->fin);
        vd->flag_open = 0;
        return -1;
    }
    goto loop;
exit_loop:;
    // set read pointer at the original point
    fseek(vd->fin, hpos, SEEK_SET); // move pointer to the original
    vd->pos = hpos;
    if (ferror(vd->fin)) {
        i = fclose(vd->fin);
        vd->flag_open = 0;
        return -1;
    }

    return 0;
}

int vdif_header_read_mode2_multi(vdif_file_t *vd)
{
    /*
        header read and analyze, then pointer not back
        (read given thread# header)
        2019-10-30
    */
    int i;
    fpos_t pos;  // debug

loop:
    //fgetpos(vd->fin, &pos);  printf("Position at 1st    %lld (%lld)\n", vd->pos,pos);  // debug

    if (fread(vd->head, sizeof(u_int32_t), vd->headlen, vd->fin) == 0) {  // header read
        i = fclose(vd->fin);
        vd->flag_open = 0;
        return -1;
    }
    vd->pos += vd->headlen * 4;
    parse_vdif_header(vd->head, &vd->vhead);   // 1st header
    //printf("       Thread #   read %d   given %d\n", vd->vhead.thread_id, vd->pup_thread); // debug
    if (vd->vhead.thread_id == vd->pup_thread) goto exit_loop;
    // forward next frame start
    vd->pos += vd->frame_bytes;
    fseek(vd->fin, vd->pos, SEEK_SET); // move pointer to the next frame 
    if (ferror(vd->fin)) {
        i = fclose(vd->fin);
        vd->flag_open = 0;
        return -1;
    }
    goto loop;
exit_loop:;
    //fgetpos(vd->fin, &pos);  printf("Position at last    %lld (%lld)\n", vd->pos,pos);  // debug

    return 0;
}



int setpos_1st_frame(vdif_file_t *vd)  // moved from vdifsch2m.c on 2019-12-11
{
    /*
        set position to 1st frame including header of pick up thread #
        2019-10-21
        2019-12-11 revised to implement skip 1st skip_bytes

    */
    int i;
    int ierr;
    ierr = 0;
    if (vd->flag_open == 0) {  // 2018-12-11
    /* VDIF file open for conversion */
        if ((vd->fin = fopen(vd->fname, "rb")) == NULL) {
            printf("setpos_1st_frame: File (%s) Open Error! \n", vd->fname);
            ierr = -1;
            goto exit2return;
        }
        vd->flag_open = 1;
    }
    // rewind file
    //rewind(vd->fin);

    // set start file position instead of rewind  2019-12-11
    fseek(vd->fin, vd->skip_bytes, SEEK_SET);      // 2019-12-11
    vd->pos = ftell(vd->fin);  // get current pointer position  2019-12-11

    // 
    while (1) {
        /* VDIF header read */
        ierr = vdif_header_read_mode1(vd);
        if (ierr < 0) goto exit2return;
        // check thread ID
        if (vd->pup_thread == vd->vhead.thread_id) break;

        // move read pointer by 1 frame
        fseek(vd->fin, vd->frame_bytes + 32, SEEK_CUR);
        if (ferror(vd->fin)) {
            i = fclose(vd->fin);
            vd->flag_open = 0;
            return -1;
        }
        vd->pos += (vd->frame_bytes + 32);
    }
exit2return:
    if (ierr < 0) {
        i = fclose(vd->fin);
        vd->flag_open = 0;
        return -1;
    }
    return 0;
}
