/**************************************************************/
/*                                                            */
/*   Program Name :  vdifcheck                                */
/*                                                            */
/*      Version:  0.01  2014-04-10                            */
/*                1.21  2019-10-31                            */
/*                                                            */
/*  Copyright (c) 2014- T.Kondo  All Right Reserved           */
/*                                                            */
/**************************************************************/
/*

  VDIF file check

          Version 2014-04-10 start development
		          2014-04-25 release Ver.1.01
		          2014-05-13 release Ver.1.10
		          2018-02-03 revised
		          2018-02-11 revised
                  2018-12-11 debug in "get_thread_num" function
                  2019-09-23 debug in "vdif_dat_anal" function
                  2019-10-12 analyzing occurrence probability by AD level
                  2019-10-31 debug in "get_thread_num" function

    How to run
      vdifcheck vdif_file [options] 
    
         where  vdif_file -- vdif file name

                 options 
                    -o[n_sec] --- set diplay just on second frame
					              (1st and last frame data are always displayed)
								  same as "-dmode 2"
                    -d[mode] 0|1|2--- set diplay mode  
					          0: 1st and last frame only (default)
							  1: every frame
							  2: just on second frame but
					             1st and last frame data are always displayed
					-f[ormat] 0|1|2|3|99
					       -- 0: Automatic format according to vdif mode edv# (default)
						      1: Compulsory set legacy format display (ED is not displayed)
							  2: Automatic format but edv# is not considered
							  3: Automatic format but edv# is set to Kashima format
                             99: debug purpose
					-n[frame] nmax  --- # of frames to be read (default all)
                    -l[egacy] --------- set legacy header length (16 bytes) compulsory (debug purpose)
					-s[kip] n --- skip 1st n bytes
					-k[ashima] [bytesframe]
                              -- set header format to Kashima compulsory 
					             bytesframe: +VE is bytes/frame set compulsory
					-a[ll_message] --- diplay detailed debug message

                    -pup thread_num  
                           -- set pick up thread#  (default 1st encountered thread)
                    -h[istogram] -- display AD level statistics
                               (show occurrence probability of AD level)



C:\IPVLBI\data\vdif\G106023100a.dat
D:\chkdata\vdif\MBL1_2014112194130.bin					
					
*/

#include "libvdif.h"

#define PROGRAM_VERSION "Ver 1.21  2019-10-31"

//#define HEAD_SEARCH_RANGE 600  // serach range (in 32bit word)  ==> move to libvdif.h
//#define VDIF_KASHIMA_SYNC 0xACABFEEDUL    // sync pattern for VDIF Kashima format ==> move to libvdif.h
#define BUFMAX 25000   // maximum read buffer size

/*
typedef struct vdif_thread_info     // VDIF thread information type   added on 2018-02-05
{
    int nthread;    // # of threads
    int id[50];     // thread ID by thread
    int frame_count; //  continuous # of frame counter during unchanged thread ID by thread
} vdif_thread_info_t;
*/

typedef struct vdifcheck_runp     // run parameters for vdifcheck
{
    char vdifname[256];  
    int mode;
	int dmode;
	int on_sec_flag;
	int format;
    int legacy;
	int flag_compulsory;  // set header format to Kashima compulsory (debug purpose)
	                      // and bytes/frame compulsory 
	int nmax;
	int all_flag; 
	int skip_byte;
    int pup_thread;    // pickup thread#  (-1 maenas use of 1st encountered thread) 2019-10-11
    int flag_hist;   // 1: analyze sampling lebel statistics  0: no
} vdifcheck_runp_t;

typedef struct vdifcheck_internal     // vdifcheck internal parameters
{
    double fbps;     // supposed data rate (bps)
	double fsample;  // supposed sampling frequency (Hz)
	double tframe;   // total number of frames (double) (computed from file size)
	double dlength;  // data length in seconds
    double pframe;   // frame length in sec (computed from headder info)

	int framsec;     // frames/sec per channel
	int badframes;   // number of invalid data frames
	int totalframes; // number of frames actually counted

    int numthread;   // # of threads  2018-02-11

    int64_t fsize;   // file size (bytes)
} vdifcheck_internal_t;

typedef struct histogram     // histgram type variable   2019-10-12
{
    unsigned int *ubuf;  // VDIF 1 frame data excepting header
    int nword;           // # of element of ubuf
    int nch;    // # of CHs
    int adbit;  // AD resolution
    int level;  // # of levels in histgram
    int64_t **ihist;    // histgram counter array  ihist[nch][level] 
    double **xhist;    // occurrence probabilty   xhist[nch][level]
    int e2bit;  // encoding method in case of 2bit AD
} histogram_t;

void param_get_vdifcheck(int argc, char *argv[], vdifcheck_runp_t *runp)                       
{
    char cbuf[256];

    // initialize parameters

    runp->mode=0;
    runp->dmode=0;
	runp->nmax=0;
	runp->format=0;
	runp->on_sec_flag=0;
    runp->vdifname[0]='\0';
    runp->legacy=0;
	runp->all_flag=0;
	runp->skip_byte=0;
	runp->flag_compulsory=0;
    runp->pup_thread = -1;  // 2019-10-11
    runp->flag_hist = 0;  // 2019-10-12

    /* 1st parameter (vdif file name) get */
    if (argc>=2){
       strcpy(runp->vdifname,*(argv+1));
    }

    /* command parameters get */
    if(cmnd_get(argc,argv,"-m",cbuf)==0) runp->mode = atoi(cbuf);      // mode (not used)
    if(cmnd_get(argc,argv,"-d",cbuf)==0) runp->dmode = atoi(cbuf);      // dmode
    if(cmnd_get(argc,argv,"-n",cbuf)==0) runp->nmax = atoi(cbuf);      // nmax
    if(cmnd_get(argc,argv,"-f",cbuf)==0) runp->format = atoi(cbuf);      // format
	if(cmnd_get(argc,argv,"-s",cbuf)==0) runp->skip_byte = atoi(cbuf);      // skip byte
    if(cmnd_get(argc,argv,"-l",cbuf)==0) runp->legacy=1;      // legacy
    if(cmnd_get(argc,argv,"-a",cbuf)==0) runp->all_flag=1;      // set all flag
    if(cmnd_get(argc,argv,"-o",cbuf)==0) runp->on_sec_flag=1;      // set on second flag
    if(cmnd_get(argc,argv,"-c",cbuf)==0 ||cmnd_get(argc,argv,"-k",cbuf)==0 
		||cmnd_get(argc,argv,"-K",cbuf)==0 ){                          // set compulsory mode
			runp->flag_compulsory=1;
        if(cbuf[0] != '\0'){
            if(cbuf[0] != '-') runp->flag_compulsory=atoi(cbuf);
        }
    }
    if(cmnd_get(argc,argv,"-pup",cbuf)==0) runp->pup_thread = atoi(cbuf);   // pick up thread#
    if(cmnd_get(argc,argv,"-h",cbuf)==0) runp->flag_hist=1; // make histgram 2019-10-12
}

void print_usage(void)
{
    printf("vdifcheck %s\n",PROGRAM_VERSION);
    printf("\n");
    printf("Usage: vdifcheck vdif_file [options]\n");
    printf("\n");
    printf("   where  vdif_file ---- vdif file name to cheak\n");
    printf("          options (any order)\n");
//    printf("             -m[ode] 0|1|2|3\n");
//	printf("                         0: Just check # of bits and time (default)\n");
//    printf("                            only error occured frame is diplayed.\n");
//    printf("                         1: Display sampling statistics (zero balance)\n");
//    printf("                         2: same as mode 0 but each frame is displayed\n");
//    printf("                         3: Display sampling statistics by AD level\n");
    printf("             -f[ormat] 0|1|2|3|99  -- set display format\n");
    printf("                         0: Automatic format according to vdif mode and edv#\n");
    printf("                            (default)\n");
    printf("                         1: Compulsory set legacy format display\n");
    printf("                            (Extended Data is not displayed)\n");
    printf("                         2: Automatic format but edv# is not considered\n");
    printf("                         3: Edv# is set to Kashima format compulsory\n");
    printf("                        99: For debug purpose\n");
//    printf("             -o[n_sec] --- set diplay just on second frame\n"); 
//    printf("					       (1st and last frame data are always displayed)\n"); 
    printf("             -d[mode] 0|1|2--- set diplay mode  \n"); 
    printf("                         0: 1st and last frame only (default)\n"); 
    printf("                         1: every frame\n"); 
    printf("                         2: just on second frame but\n"); 
    printf("                            1st and last frame data are always displayed\n"); 
//    printf("             -a[ll] --- all data check mode\n");
	printf("             -n[frame] nmax  --- # of frames to be read (default all)\n");
    printf("             -l[egacy] -- legacy header length (16 bytes) compulsory\n"); 
    printf("             -s[kip] n -- skip 1st n bytes\n");
    printf("             -k[ashima] [bytesframe]\n");
    printf("                       -- set header format to Kashima compulsory \n");
	printf("                          bytesframe: +VE is bytes/frame set compulsory\n");
    printf("             -pup thread_num\n");                   // 2019-10-11
    printf("                       -- set pick up thread# \n"); // 2019-10-11
    printf("             -h[istogram]\n");    // 2019-10-12
    printf("                       -- display AD level statistics \n");
	printf("                          occurrence probabilities by sampling level\n");
    printf("\n");
}

void headerprint_for_debug(unsigned int *head, int headlen, vdif_header_t *vhead)
{
	int n;
	printf("\n");
	printf("========================= VDIF HEADER INFORMATION =========================\n");
	printf(" VDIF header (raw)        :");
	for(n=0; n<4; n++) printf(" %08X",head[n]); printf("\n");
	if(headlen > 4){
	printf("    extended header (raw) :");
	for(n=4; n<headlen; n++) printf(" %08X",head[n]); printf("\n");
	}
	printf("\n");
	printf(" Invalid flag = %d :", vhead->invalid);
	printf("       Legacy flag = %d\n", vhead->legacy);
	printf(" Sec from ref epoch = %d :", vhead->sec_fm_ref);
	printf("  Ref epoch = %d", vhead->ref_epo);
	printf("   (%04d/%02d/%02d %02d:%02d:%02d)\n", 
		   vhead->t.year,vhead->t.month,vhead->t.mday,vhead->t.hh,vhead->t.mm,vhead->t.ss);

	printf(" Frame # = %d :", vhead->frame);
	printf("  VDIF Ver# = %d :", vhead->ver);
	printf(" # Chs(log2) = %d\n", vhead->log2_chs);
	printf(" Frame length in 8 byte unit = %d  (= %d bytes)\n", vhead->frame_len64,vhead->frame_len64*8);
	printf(" Data type = %d :", vhead->d_type);
	printf(" #bits/sample-1 = %d :", vhead->bpsampl);
	printf(" Thread ID = %d\n", vhead->thread_id);
	printf(" Station ID (A2) = %s ", vhead->sid);
	printf("   ( = %d in number)\n", vhead->station_id);
	printf(" EDV : %d\n", vhead->edv);
	printf(" Exteded Data W1 W2 W3 W4 : %06X %08X %08X %08X\n", vhead->w1,vhead->w2,vhead->w3,vhead->w4);

	printf(" --- in case of NICT Extended Format ---\n");	   
	//printf(" Head[4] = %08X :", vhead->w1);
	printf("  uflag = %d :", vhead->uflag);
	printf("  srate = %d :", vhead->srate);
	printf("  Sync block = %08X\n", vhead->sync_block);
	printf("  DAS/station name (A8) : %s\n", vhead->das);
	printf("===========================================================================\n");


}

void headerprint(vdif_header_t *vhead)
{
	printf("   Time        : %04d/%02d/%02d %02d:%02d:%02d  Frame# : %d\n", 
		   vhead->t.year,vhead->t.month,vhead->t.mday,vhead->t.hh,vhead->t.mm,vhead->t.ss,
		   vhead->frame);
	printf("   Bytes/Frame : %d bytes\n", vhead->frame_len64*8);
	if(vhead->d_type==0){
	    printf("   Data type   : Real Data    ");
    } else {
	    printf("   Data type   : Complex Data ");
    }
    printf("        AD bits : %d", vhead->bpsampl+1);
    printf("    #Channels : %d\n", (int)pow(2.0, vhead->log2_chs));
	printf("   EDV#        : %d\n",vhead->edv);
	if(vhead->edv==1){   // This is NICT's EDV
       printf("   Extended header information (NICT\'s EDV)\n");
	   printf("      Sampling Frequency : %d", vhead->srate);
	   if(vhead->uflag==0){
		   printf(" kHz\n");
	   } else {
		   printf(" MHz\n");
	   }
	   printf("      Sync block         : %08X\n", vhead->sync_block);
	   printf("      DAS/Station name   : %s\n", vhead->das);
	}
}


void disp_summary(vdifcheck_runp_t *runp, vdif_header_t *vhead1st, vdif_header_t *vheadlast,
				   vdifcheck_internal_t *ip)
{
/*
       summary display
*/

	printf("\n");
	printf("========================= SUMMARY =============================\n");
	printf(" File Name   : %s\n",runp->vdifname);
	printf("      Size   : %lld bytes\n",(long long)ip->fsize); 
    printf(" # of threads: %d\n",ip->numthread);  // 2018-02-11 
	printf(" Start Time  : %04d/%02d/%02d %02d:%02d:%02d\n", 
		   vhead1st->t.year,vhead1st->t.month,vhead1st->t.mday,vhead1st->t.hh,vhead1st->t.mm,vhead1st->t.ss);
	printf("               Thread ID : %d      Frame# : %d\n", vhead1st->thread_id,vhead1st->frame);
	printf(" Bytes/Frame : %d bytes\n", vhead1st->frame_len64*8);
	if(vhead1st->d_type==0){
	    printf(" Data type   : Real Data    ");
    } else {
	    printf(" Data type   : Complex Data ");
    }
    printf("    AD bits : %d", vhead1st->bpsampl+1);
    printf("    #Channels : %d\n", (int)pow(2.0, vhead1st->log2_chs));
	printf(" EDV#        : %d\n",vhead1st->edv);
	if(vhead1st->edv==1){   // This is NICT's EDV
       printf(" Extended header information (NICT\'s EDV)\n");
	   printf("      Sampling Frequency : %d", vhead1st->srate);
	   if(vhead1st->uflag==0){
		   printf(" kHz\n");
	   } else {
		   printf(" MHz\n");
	   }
	   printf("      Sync block         : %08X\n", vhead1st->sync_block);
	   printf("      DAS/Station name   : %s\n", vhead1st->das);
	}
	printf(" # of frames/sec     : %d\n",ip->framsec);     
	printf(" # of total frames   : %d",ip->totalframes);     
	printf("   (#bad frames : %d)\n",ip->badframes);     
	//printf(" Data period (sec)   : %f\n",ip->dlength);     
    printf(" Data period (sec)/thread : %f\n",ip->dlength/(double)ip->numthread);  // 2018-02-11     
	printf(" Data Rate/thread (Mbps)  : %f\n",ip->fbps/1.0e6);
	//printf(" Sampling Frequency Estimated (MHz) : %f\n",ip->fsample/1.0e6); 
	if(vhead1st->d_type==0){
	   printf(" Sampling Frequency Estimated (MHz) for real : %f\n",ip->fsample/1.0e6);      // 2018-02-11
    } else {
	  printf(" Sampling Frequency Estimated (MHz) for complex : %f\n",0.5*ip->fsample/1.0e6);    // 2018-02-11 
    }
	printf("================================================================\n");

	printf("\n");

}


int get_thread_num(vdifcheck_runp_t *runp, int *numthread, int *table_thread, int *table_frames)
{
    // 2018-12-11
    // 2019-10-31  debug

    int n;
    vdif_file_t vd;

    strcpy(vd.fname, runp->vdifname);

    vd.flag_compulsory=runp->flag_compulsory;  // 2018-12-11
    vd.skip_bytes = runp->skip_byte;  // 2019-10-31

    n=vdif_thread_anal(&vd, numthread, table_thread, table_frames);

    return n;
}

void hist_malloc(histogram_t *ht)
{
    // 2019-10-12
    int n;
    ht->ubuf = (unsigned int *)malloc(ht->nword * sizeof(unsigned int));  // 2019-10-11
    ht->ihist = malloc(ht->nch*ht->level * sizeof(int64_t *));
    ht->xhist = malloc(ht->nch*ht->level * sizeof(double *));
    for (n = 0; n < ht->nch; n++) {
        ht->ihist[n] = malloc(ht->level * sizeof(int64_t));
        ht->xhist[n] = malloc(ht->level * sizeof(double));
    }
}

void hist_free(histogram_t *ht)
{
    // 2019-10-12
    int n;
    free(ht->ubuf);
    for (n = 0; n < ht->nch; n++) {
        free(ht->ihist[n]);
        free(ht->xhist[n]);
    }
    free(ht->ihist);
    free(ht->xhist);
}

void hist_reset(histogram_t *ht)
{
    // 2019-10-12
    int n, k;
    for (n = 0; n< ht->nch; n++) {
        for (k = 0; k < ht->level; k++) {
            ht->ihist[n][k] = 0;
        }
    }

}

void hist_calc_prob(histogram_t *ht)
{
    // 2019-10-12
    int n, k;
    double total;
    for (n = 0; n<ht->nch; n++) {
        total = 0.0;
        for (k = 0; k < ht->level; k++) total += (double)ht->ihist[n][k];
        if (total > 0.0) {
            for (k = 0; k < ht->level; k++) ht->xhist[n][k] = (double)ht->ihist[n][k] / total;
        }
        else {
            for (k = 0; k < ht->level; k++) ht->xhist[n][k] = 0.0;
        }
    }

}


void hist_anal(histogram_t *ht)
{
/*
      analyze level distribution
      2019-10-12

*/
    int iw;
    int bitstream;
    int n, i, k, ww, pat, wd;
    int ival;
    int ikai;
    int chn, m;
    int pattbl[9] = { 0x00,0x01,0x03,0x07,0x0F,0x1F,0x3F,0x7F,0xFF };  // bit mask pattern for AD resolution

    int *conv2bit;
    int conv2bit_1[] = { 0,1,2,3 };   // offset binary (no conversion 2 bit)
    int conv2bit_2[] = { 0,2,1,3 };   //  conversion 2 bit (default)  <= Mark5B
    int conv2bit_3[] = { 1,0,2,3 };   //  conversion 2 bit singed integer

    conv2bit = conv2bit_1;     // VDIF default
    if (ht->e2bit == 1) conv2bit = conv2bit_1;
    if (ht->e2bit == 2) conv2bit = conv2bit_2;
    if (ht->e2bit == 3) conv2bit = conv2bit_3;

    bitstream = ht->nch * ht->adbit;
    ikai = 32 / bitstream;     // # of independent time data
    pat = pattbl[ht->adbit];

    for (m = 0; m < ht->nch; m++) {
        chn = m+1;
        k = 0;
        for (n = 0; n < ht->nword; n++) {
            wd = ht->ubuf[n];
            wd >>= (chn - 1)*ht->adbit;
            for (i = 0; i < ikai; i++) {
                ww = (int)(wd & pat);
                if (ht->adbit == 2) {
                    ww = conv2bit[ww];
                }
                ht->ihist[m][ww]++;   
                wd >>= bitstream;
                k++;
            }
        }
    }

}

void hist_summary(histogram_t *ht)
{
    // print summary of level distribution
    // 2019-10-12
    int n, k;
    hist_calc_prob(ht);

    printf("\n");
    if (ht->nch <= 8) {
        printf(" ============ PROBABILITY(%%) OF OCCURRENCES ============\n");
    }
    else {
        printf(" ===================== PROBABILITY(%%) OF OCCURRENCES ==================\n");
    }
    printf("  LEVEL");
    if (ht->nch <= 8) {
        for (n = 0; n < ht->nch; n++) printf("  CH%02d", n + 1); printf("\n");
        printf(" -------------------------------------------------------\n");
        for (k = 0; k < ht->level; k++) {
            printf(" %4d  ", k);
            for (n = 0; n < ht->nch; n++) printf("%6.1f", ht->xhist[n][k] * 100.0);
            printf("\n");
        }
        printf(" =======================================================\n");
    }
    else {
        for (n = 0; n < ht->nch; n++) printf(" CH%02d", n + 1); printf("\n");
        printf(" ----------------------------------------------------------------------\n");
        for (k = 0; k < ht->level; k++) {
            printf(" %4d ", k);
            for (n = 0; n < ht->nch; n++) printf("%5.1f", ht->xhist[n][k] * 100.0);
            printf("\n");
        }
        printf(" ======================================================================\n");
    }
    //hist_reset(ht);
}

void vdif_dat_anal(vdifcheck_runp_t *runp, vdif_header_t *vhead1st, vdif_header_t *vheadlast,
				   vdifcheck_internal_t *ip)
{
/*
     VDIF data file analysis   2014-04-15
                               2018-02-05
                               2019-09-23 debug
    [in]
	  vdifname  ---  VDIF data file name
      nmax     ---- Maximum frame # to read  ( if 0, all data are read)
                    or Maximum seconds for every second display mode

	[out]
	  vhead1st  ----- 1st VDIF header
	  vheadlast ----- last VDIF header
	  fbps      ----- Supposed data rate (bps)
	  fsps      ----- Supposed sampling frequency (Hz)
      framsec   ----- frames/sec per channel
      fsize     ----- file size
      tframe    ----- total number of frames (double) (computed from file size)
      pframe    ----- period of frame (computed from headder info)
      dlength   ----- data length in seconds
	  badframes ----- number of invalid data frames
	  totalframes --- number of frames actually counted


*/
       int i,p_flag;
       int n, k;
       unsigned int syncpat=VDIF_KASHIMA_SYNC;    // sync pattern for VDIF Kashima format
       unsigned int ibuf[2504];  // for 1 Frame data 2504 X 32 bits
	   unsigned int *head;   // for header 4 X 32 bits
	   unsigned int *idata;  // for data field 2500 X 32 bits
	   unsigned int head1st[8];
       int frame32=2500;   // # of 32-bit words in a Mark5B frame
       int nbuf=2504;   //  frame length including header (in 32-bit word)
       int iread; 
	   int sec_old;
	   int headlen;
	   int iok;
       int frame_c; // frame # counter
       //int data1st[2]; // First 2 word data just after headder at given bit position
	   int nfram_per_sec;  //  true # of frames per sec
	   int nfram_per_sec_flag;  //  true # of frames per sec count OK flag
	   int edv;
	   int legacy_mode;
	   int form_mode;
	   int ref_thread_id; // reference thread id to calculate true # of frames in second
       int nch,adbit;
	   char sunit[10];
	   char cbuf[10];
       double fsample;
       int64_t dfsize;
	   //int64_t frame_bytes;
	   int64_t bytes_frame;  // bytes/frame (including header)
	   int64_t b_frame_c;    // bytes per frame in data  (true header to header)
	   int64_t b_frame_c_old;
	   int64_t pos,wpos, wpos_old;
       vdif_header_t vhead;
       FILE *fin;
       int numthread;                     // 2018-02-03
       int table_thread[50];              // 2018-02-03
       int table_frames[50];              // 2018-02-05
       //unsigned int *ubuf;  // for 1 frame data 2019-10-11
       int wordread; // 2019-101-11
       //int nword, wordread; // 2019-101-11
       //int **xhist;    // histgram table  xhist[nch][level]  2019-10-12
       //int level;      // 2019-10-12
       histogram_t ht;

       //mk5_aux_t aux;


       printf("Data File : %s\n",runp->vdifname);
       // thread check

       printf("  now analyzing # of threads ...... ");
       if(get_thread_num(runp, &numthread, table_thread, table_frames) < 0) exit(0);    // 2018-02-03
       printf("\n");
       printf("  # of threads included is %d\n",numthread);              // 2018-02-03
       if(numthread==1){                                                 // 2018-02-03
           printf("  thread ID   : %d \n",table_thread[0]);              // 2018-02-03
           printf("  # of frames in a second  : %d\n",table_frames[0]);  
       } else {                                                          // 2018-02-03
           printf("  thread IDs  : ");                                   // 2018-02-03
           for(n=0; n<numthread; n++) printf(" %d",table_thread[n]);     // 2018-02-03
           printf("\n");                                                 // 2018-02-03
           for(n=0; n<numthread; n++){
              printf("  thread ID and # of frames continuing same thread ID : %d %d\n",table_thread[n],table_frames[n]);
           }
       }        
                                                         // 2018-02-03

       ip->numthread=numthread;  // 2018-02-11


	   headlen=8; // default header length (in 32-bit word)
	   head=ibuf;
	   idata=ibuf+4;

	   nfram_per_sec=0; 
	   nfram_per_sec_flag=0;

	   form_mode=0;     //  1:  Legacy style display   2: Kashima edv   3: versatile edv   

	   if(runp->on_sec_flag==0){
	      if(runp->dmode==2) {
		      runp->on_sec_flag=1;
	      } 
	   } else {
		   runp->dmode=2;
	   }

	   //ref_thread_id=0;
       //ref_thread_id=table_thread[0];   // 2018-02-05  deleted on 2019-10-11
       if (runp->pup_thread < 0) {            // 2019-10-11
           ref_thread_id = table_thread[0];   // 2019-10-11
       }                                      // 2019-10-11
       else {                                 // 2019-10-11
           ref_thread_id = runp->pup_thread;  // 2019-10-11
       }                                      // 2019-10-11

	   iok=0;
       /* Data file open */
	   if( (fin  = fopen( runp->vdifname, "rb" )) == NULL )
       {
           printf("Data File %s   Open Error! \n",runp->vdifname);
		   iok=-1;
		   exit(1);
       }
       //printf("Data File : %s\n",runp->vdifname);
       dfsize=get_file_size(runp->vdifname);
       ip->fsize=dfsize;

	   if(runp->skip_byte > 0) {  // skip 1st skip_byte
		   printf(" Skipping 1st %d bytes .....",runp->skip_byte);
		   fseek(fin,runp->skip_byte,SEEK_SET);
		   printf("\n");
	   }

	   // check 1st portion is a possible VDIF header or not 
	   if(check_vdif_header(fin) < 0){
		   printf(" 1st portion is not a VDIF header.\n");
		   printf(" You want to search header assuming VDIF Kashima format? ([y]/n)");
           read1line(cbuf);
		   if(cbuf[0]=='n' || cbuf[0]=='N'){ // yes
			   exit(1);
		   } else {
			   runp->flag_compulsory=1;
		   }
	   }

	   if(runp->flag_compulsory > 0){
		   if(find_Kas_sync(fin, &pos, 1)<0){
              printf("Data File does not have VDIF Kashima header!\n");
		      iok=-1;
		      exit(1);
		   }
		   printf(" Data file has VDIF Kashima header.\n");
		   if(pos > 0){
			   printf(" But 1st header is not a correct header!!\n");
		       printf(" 1st correct header is located at %lld byte position\n",(long long)pos);
		   }
	   }

	   // header read

	   if(runp->flag_compulsory > 0) headlen=8;

	   while(1){
	      wpos=ftell(fin);  // save current pointer position

          //iread = fread( ibuf, sizeof( unsigned int ), headlen, fin );  // 1 frame read
          iread = fread( ibuf, sizeof( unsigned int ), headlen+2, fin );  // 1 frame headder + data 2 32bit word
          if (iread ==0 ) { // EOF
	         printf("\n");
             goto exitreturn; 
          }
          parse_vdif_header(head, &vhead);

	      memcpy(head1st, head, headlen*4);  // keep 1st header 
	      *vhead1st=vhead;                   // keep 1st header (analyzed)

	      if(runp->flag_compulsory > 0) {
		     if(vhead.w2 != syncpat){
		        if(find_Kas_sync(fin, &pos, 1)<0){
                    printf("Data File does not have VDIF Kashima header!\n");
		            iok=-1;
		            exit(1);
		        }
		        if(pos > 0){
		            printf(" correct VDIF Kashima header found at %lld byte position\n",(long long)pos);
					wpos=pos;
		        }
			 } else {
			    goto exit1;
			 }
		  } else {
		     goto exit1;
		  }

	   }

exit1:

       //printf("Debug thread ID = %d\n",vhead.thread_id);
	   //ref_thread_id=vhead.thread_id;

       printf("  Thread ID to pick up ==> %d\n",ref_thread_id); // 2018-02-05

       strcpy(sunit,"(kHz)");                             // this is for Kashima extended
       if(vhead.uflag==1) strcpy(sunit,"(MHz)");          //   

	   if(runp->flag_compulsory > 1) vhead.frame_len64=runp->flag_compulsory/8;

	   ip->tframe=(double)dfsize/(double)(vhead.frame_len64*8);

	   printf(" 1st header information is as follows\n");
	   headerprint_for_debug(head1st,headlen,vhead1st);
	   printf("\n");

	   // VDIF header length re setup in case of legacy mode
	   if(runp->legacy==1) headlen=4;
	   if(vhead.legacy==1) headlen=4;  // in 32-bit word


	   if(runp->flag_compulsory > 0) headlen=8;



	   //if (((vhead.frame_len64*2-headlen) % frame32) !=0) {
	   //    printf("Warning!!\n");
	   //} else {
	   //   nloop=(vhead.frame_len64*2-headlen)/frame32;
	   //}

	   // set read pointer 
       //fseek(fin,0L,SEEK_SET); // move pointer to the 1st of file
       fseek(fin,wpos,SEEK_SET); // move pointer to the beginning of headder
       if( ferror(fin) ){
          goto exitreturn; 
       }

	   bytes_frame=vhead.frame_len64*8;    // supposed frame bytes including header
	   //frame_bytes=(vhead.frame_len64*2-headlen)*4;    // supposed frame bytes excepting header
	   //frame_bytes=(vhead.frame_len*2-headlen)*8+headlen*4;  // true frame byte excepting header

       /** added on 2019-10-11 **/
       nch = (int)pow(2.0, vhead1st->log2_chs);  // # of channels   2019-10-12 moved from last portion
       adbit = vhead1st->bpsampl + 1;             // AD bits   2019-10-12 moved from last portion
       ht.nword = (bytes_frame/4)-headlen;   // 2019-10-11
       ht.nch = nch;
       ht.adbit = adbit;
       ht.level = (int)pow(2.0, adbit);
       hist_malloc(&ht); 
       hist_reset(&ht);
       /** add end **/

	   // initialize counter
	   frame_c=0;
	   ip->badframes=0;
	   sec_old=vhead.sec_fm_ref;
       b_frame_c = 0;  // 2019-09-23
	   // set Extended format
	   edv=vhead.edv;
	   if(runp->flag_compulsory > 0) {
		   edv=1; // compulsory set Kashima mode
		   //if(runp->flag_compulsory >1) frame_bytes=runp->flag_compulsory-headlen*4;
		   if(runp->flag_compulsory >1) bytes_frame=runp->flag_compulsory;
	   }

	   legacy_mode=vhead.legacy;
	   //legacy_mode=1;  // debug


	   // set display format
	   switch(runp->format){
		   case 0:       // automatic mode
			       if(vhead.legacy==1){
			          form_mode=0;      // legacy
		           } else {
			          if(vhead.edv==1){
				          form_mode=1;  // with Kashima edv
			          } else {
				          form_mode=2;  // versatile edv
			          }
		           }
				   break;
		   case 1:       // compulsory set legacy mode
			       form_mode=0;
				   break;
		   case 2:       // automatic but versatile EDV format
			       if(vhead.legacy==1){
			          form_mode=0;      // legacy
		           } else {
				      form_mode=2;  // versatile edv
		           }
				   break;
		   case 3:     // automatic but Kashima EDV format
			       if(vhead.legacy==1){
			          form_mode=0;      // legacy
		           } else {
				      form_mode=1;  // Kashima edv
		           }
				   break;
		   default: form_mode=0;

	   }
	   if(runp->flag_compulsory > 0) form_mode=1;
	   if(runp->format==99) form_mode=99;   // debug


       printf(" ==============================================================================\n");
	   switch(form_mode){
		   case 0:   // legacy
		       printf(" HH:MM:SS  FRAME#  I  L  VER  CHS     #Bytes  C AD  TID SID SEC_FM_EPO  EPOCH\n");
		       break;
		   case 1:   // with Kashima ED
		       printf("                                                    Sampling\n");
	           printf(" HH:MM:SS  FRAME#  I  L  VER  CHS     #Bytes  TID SID  %s    SYNC   DAS/Stat\n",sunit);
			   break;
		   case 2:   // with Versatile ED
		       printf("                                                             EXTENDED DATA\n");
	           printf(" HH:MM:SS  FRAME#  I  L  VER  CHS     #Bytes  C AD  TID SID       W1       W2\n");
			   break;
		   case 99:   // for debug
		       printf("                                 EXTENDED             DATA\n");
			   printf(" HH:MM:SS  FRAME#    #Bytes     W1       W2       D1       D2    TotalFRAM#\n");
			   break;
	   }
       printf(" ------------------------------------------------------------------------------\n");

       while(1){

           /************************/
           /*  Entry of Frame loop */
           /************************/

		   while(1){
	         wpos=ftell(fin);  // save current pointer position

             iread = fread( ibuf, sizeof( unsigned int ), headlen+2, fin );  // 1 frame headder + data 2 32bit word
             if (iread ==0 ) { // EOF
	            printf("\n");
                goto exithere; 
             }
             parse_vdif_header(head, &vhead);

       //printf("Debug thread ID = %d\n",vhead.thread_id);

			 // check header shync block in case of compulsory set Kashima mode
	         //if(runp->flag_compulsory > 0) {
	         if(edv==1) {
		        if(vhead.w2 != syncpat){
		           if(find_Kas_sync(fin, &pos, 0)<0){  // header search with mode=0
                       printf("VDIF Kashima header can not be found in %d words!\n",HEAD_SEARCH_RANGE);
					   printf("  search header further? ([y]/n) ");
                       read1line(cbuf);
		               if(cbuf[0]=='n' || cbuf[0]=='N'){ // yes
		                   if(find_Kas_sync(fin, &pos, 1)<0){  // header search with mode=1
                               printf("VDIF Kashima header can not be found after all\n");
		                       iok=-1;
		                       goto exithere;
						   }
					   } else {
		                   iok=-1;
		                   goto exithere;
					   }
		           }
		           if(pos > 0){
					   if(runp->all_flag==1) printf(" correct VDIF Kashima header found at %lld byte position\n",(long long)pos);
					   wpos=pos;
		           }
			    } else {
			       goto exit2;
			    }
		     } else {
		        goto exit2;
		     }

	      }

exit2:
          frame_c++;  // total frame number counter

		  if(frame_c > 1){
              b_frame_c = wpos-wpos_old;   // frame length (in bytes)
		  }

		  if(vhead.sec_fm_ref!=sec_old) {  // new second start
			  if(nfram_per_sec_flag==1) nfram_per_sec_flag=0;
		  }

		  if(runp->on_sec_flag==1){
		     if(frame_c>0){
			   if(vhead.sec_fm_ref!=sec_old) p_flag=1;  // new second start
			 }
			 if(frame_c==1) p_flag=1;  // to display 1 st frame data
		  } else {
			  p_flag=1;
			  if(runp->dmode==0 && frame_c > 1) p_flag=0;
		  }


		  // check frame length difference
		  if(frame_c > 1){
			  //if(b_frame_c != frame_bytes+headlen*4){ // frame length differs from definition
			  if(b_frame_c != bytes_frame){ // frame length differs from definition
			     p_flag=1;     // set print flag
			     printf(" Bytes/frame (%lld) differ occurred at %d-th frame\n",(long long)b_frame_c,frame_c);
                 if(vhead.invalid==0) ip->badframes++;
			  }
		  }

		  wpos_old=wpos;
		  b_frame_c_old = b_frame_c;
		  
		  if(vhead.invalid==1) ip->badframes++;

		  // #frames counts
		  if(vhead.frame==0){
			  if(vhead.thread_id==ref_thread_id) {
				  if(nfram_per_sec_flag==0 && nfram_per_sec==0) nfram_per_sec_flag=1;
			  }
		  }
		  if(nfram_per_sec_flag==1 && vhead.thread_id==ref_thread_id) nfram_per_sec++;


		  if(p_flag==1){  // print

		     printf(" %02d:%02d:%02d %6d",vhead.t.hh,vhead.t.mm,vhead.t.ss, vhead.frame);
//			 printf("  %08X\n",vhead.sync_block);  // debug
//			 if(vhead.sync_block != 0xACABFEED) {
//			 ////if(vhead.w2 != 0xACABFEED) {
//			 	 continue;
//			 }
			 
			 switch(form_mode){
				 case 0:
				     printf(" %3d %2d  %3d   %2d %10d  %1d %2d %4d  %s %10d %6d\n", 
						 vhead.invalid,vhead.legacy,vhead.ver,
					     (int)pow(2.0, vhead.log2_chs),vhead.frame_len64*8,
					     vhead.d_type,vhead.bpsampl+1,vhead.thread_id,vhead.sid,vhead.sec_fm_ref,vhead.ref_epo);
					 break;
				 case 1:
				     printf(" %3d %2d  %3d   %2d %10d %4d  %s %5d  %08X  %s\n",
					     vhead.invalid,vhead.legacy,vhead.ver,
					     (int)pow(2.0, vhead.log2_chs),vhead.frame_len64*8,vhead.thread_id,vhead.sid,
					     vhead.srate, vhead.sync_block,vhead.das);
					 break;
				 case 2:
				     printf(" %3d %2d  %3d   %2d %10d  %1d %2d %4d  %s %08X %08X\n", 
						 vhead.invalid,vhead.legacy,vhead.ver,
					     (int)pow(2.0, vhead.log2_chs),vhead.frame_len64*8,
					     vhead.d_type,vhead.bpsampl+1,vhead.thread_id,vhead.sid,vhead.w1,vhead.w2);
					 break;
				 case 99:
				     printf(" %10d  %08X %08X %08X %08X  %8d\n", 
						 vhead.frame_len64*8,vhead.w1,vhead.w2, ibuf[headlen], ibuf[headlen+1], frame_c);
					 break;
			 }
			 //if(runp->dmode==0 && frame_c==1) printf("    ----------\n");
			 if(runp->dmode==0 && frame_c==1) printf("    ...............\n");
		  }

          // koko  either 1 frame read or just skip 1frame
          if (runp->flag_hist==0) {     // just skip 1 frame 2019-10-11
              fseek(fin, wpos + bytes_frame, SEEK_SET); // pointer set to next frame
              //fseek(fin,wpos+frame_bytes+headlen*4,SEEK_SET); // pointer set to next frame
              //_fseeki64(fin,wpos+frame_bytes+headlen*4,SEEK_SET); // pointer set to next frame (VC++)
              if( ferror(fin) ){
                 goto exithere; 
              }

          }
          else {                   // read 1 frame   2019-10-11
              fseek(fin, wpos + headlen*4, SEEK_SET); // pointer set to end of header
              wordread = fread(ht.ubuf, sizeof(unsigned int), ht.nword, fin);  // 2019-10-11
              if (wordread == 0) { // EOF                                // 2019-10-11
                  goto exithere;                                         // 2019-10-11
              }                                                          // 2019-10-11
              // occurrence analysis by level
              hist_anal(&ht);  // 2019-10-11
          }

		  p_flag=0;
		  sec_old=vhead.sec_fm_ref;
		  if(runp->nmax > 0 && frame_c > runp->nmax) goto exithere;
	   }
exithere:

	   if(runp->dmode!=1 && p_flag==0){ // display last frame data
		     printf(" %02d:%02d:%02d %6d",vhead.t.hh,vhead.t.mm,vhead.t.ss, vhead.frame);
			 switch(form_mode){
				 case 0:
				     printf(" %3d %2d  %3d   %2d %10d  %1d %2d %4d  %s %10d %6d\n", 
						 vhead.invalid,vhead.legacy,vhead.ver,
					     (int)pow(2.0, vhead.log2_chs),vhead.frame_len64*8,
					     vhead.d_type,vhead.bpsampl+1,vhead.thread_id,vhead.sid,vhead.sec_fm_ref,vhead.ref_epo);
					 break;
				 case 1:
				     printf(" %3d %2d  %3d   %2d %10d %4d  %s %5d  %08X  %s\n",
					     vhead.invalid,vhead.legacy,vhead.ver,
					     (int)pow(2.0, vhead.log2_chs),vhead.frame_len64*8,vhead.thread_id,vhead.sid,
					     vhead.srate, vhead.sync_block,vhead.das);
					 break;
				 case 2:
				     printf(" %3d %2d  %3d   %2d %10d  %1d %2d %4d  %s %08X %08X\n", 
						 vhead.invalid,vhead.legacy,vhead.ver,
					     (int)pow(2.0, vhead.log2_chs),vhead.frame_len64*8,
					     vhead.d_type,vhead.bpsampl+1,vhead.thread_id,vhead.sid,vhead.w1,vhead.w2);
					 break;
				 case 99:
				     printf(" %10d  %08X %08X %08X %08X  %8d\n", 
						 vhead.frame_len64*8,vhead.w1,vhead.w2, ibuf[headlen], ibuf[headlen+1], frame_c);
					 break;
			 }
	   }
       printf(" ==============================================================================\n");

	   //memcpy(vheadlast, &vhead, sizeof(vdif_header_t));
	   *vheadlast=vhead;
	   //printf(" last header information\n");
       //headerprint(vheadlast);

  //     nch=(int)pow(2.0, vhead1st->log2_chs);  // # of channels  2019-10-12 moved to far avobe
  //     adbit=vhead1st->bpsampl+1;             // AD bits  2019-10-12 moved to far avobe
	   
	   ip->totalframes=frame_c;    // total frames all channels 
	   ip->framsec=nfram_per_sec;  // frame/sec per channel
	   ip->fbps=(double)nfram_per_sec*(double)(bytes_frame-headlen*4)*8.0;  // data rate per all channel
	   ip->fsample=ip->fbps/(double)(adbit*nch);

       if(vhead1st->edv==1){ // kashima format
           if(vhead1st->uflag==0){
               fsample=(double)vhead1st->srate*1000.0;
           } else {
               fsample=(double)vhead1st->srate*1.0e6;
           }
           ip->pframe=(double)(vhead1st->frame_len64*8-headlen*4)*8.0/((double)(nch*adbit)*fsample);
       } else {
           ip->pframe=1.0/(double)ip->framsec;
       }

       if(vhead1st->edv==1){ // kashima format
           ip->dlength=ip->pframe*(double)ip->fsize/(double)(vhead1st->frame_len64*8);
       } else {
  	      ip->dlength=(double)frame_c/(double)nfram_per_sec;
       }

       // print summary of histgram 
       if (runp->flag_hist == 1) {  // 2019-10-12
           hist_summary(&ht);
       }

 exitreturn:
       hist_free(&ht);  // 2019-10-12
       i=fclose(fin);

}



int main(int argc, char *argv[])
{
       //int validtrk[64], trktable[64], hstack_tbl[64]; 
       int mode;
       char vdifname[256];
       double telapsed;           // to measure processing time
       time_t tstart, tfinish;    //           "
       vdifcheck_runp_t runp;
	   vdifcheck_internal_t ip;  
       vdif_header_t vhead1st, vheadlast;
       vdif_file_t vd;    // debug

       if(argc == 1 || (argc == 2 && strncmp(*(argv+1),"--help",6)==0 )){ // 2008-02-19
          print_usage();
          exit(0);
       }
       if(argc == 2 && strncmp(*(argv+1),"--ver",5)==0 ){ // 2008-02-19
          printf("\n");
          printf("vdifcheck %s\n",PROGRAM_VERSION);
          printf("\n");
          exit(0);
       } 

       param_get_vdifcheck(argc, argv, &runp);                       
       /* parameters set up */
       mode=runp.mode;
       strcpy(vdifname,runp.vdifname);

	   //strcpy(vd.fname,runp.vdifname);   // debug
       //if(get_real_vdif_info(&vd)<0) return 0;  // debug

       /* Program Name Display */
       printf("***********************************************************\n");
       printf("*    VDIF data check                                      *\n");
       //printf("*      %s  by  T.KONDO/NICT             *\n",PROGRAM_VERSION);
       printf("*      %s  by  T.KONDO                  *\n",PROGRAM_VERSION);
       printf("*                                                         *\n");
       printf("***********************************************************\n");
       printf("\n");

       tstart= time(NULL);


       /* analyze VDIF data file */
       vdif_dat_anal(&runp, &vhead1st,&vheadlast, &ip);

/*
	   printf("\n");
	   printf(" 1st header information\n");
       headerprint(&vhead1st);
	   printf("\n");
	   printf(" Last header information\n");
       headerprint(&vheadlast);

	   printf("\n");

       printf(" File Size   : %lld bytes\n",(long long)ip.fsize);     
	   printf(" #frames/sec : %d\n",ip.framsec);     
	   printf(" #frames     : %d\n",ip.totalframes);     
	   printf(" #bad frames : %d\n",ip.badframes);     
	   printf(" Data period (sec)   : %f\n",ip.dlength);     
	   printf(" Data Rate/ch (Mbps) : %f\n",ip.fbps/1.0e6);     
	   printf(" Sampling Frequency (MHz) : %f\n",ip.fsample/1.0e6);     

	   printf("\n");
*/

	   disp_summary(&runp,&vhead1st,&vheadlast,&ip);
       tfinish= time(NULL);
       telapsed = difftime(tfinish,tstart);
       //printf("Time elapsed (sec) for One obs process is %f\n",telapsed);

       strcpy(vd.fname,runp.vdifname);

       //get_real_vdif_info(&vd);



       return 0;
}
