/*********************************************************************/
/*                                                                   */
/*   File name :  vdifsch2m.h                                        */
/*                                                                   */
/*      Version:   2019-11-10                                        */
/*                                                                   */
/*   Copyright (c) 2019-    T.Kondo    All Right Reserved            */
/*                                                                   */
/*********************************************************************/
/*



2019-10-24 
     start development
2019-11-10
     fsample_orig is added in runp_t type variable
*/


#include "libvdif.h"


typedef struct runp     // run parameters for vdif to Mark5B conversion program
{
    int ptype;   // run parameter type  =1  mode 1,  =2 mode 2
    int channel;
    int group;
    int unit;
    int soffset;
    int period;
    int reverse;
	int64_t skip_bytes;   // skip bytes (1st skip over # of bytes)
    int scan;    // scan# for get MODE from VEX file (default=1) 
    int odd;    // =1 4byte integer offset mode in case of 64 track
    int monit;   // =1  inofo_file monitor ON,  =0 no monitoring
    char vdifname[256];   // 2007-07-17  80->256
    //char k5name[256]; // 2007-07-17  80->256
    char oname[256];
    char oname_given[256]; // 2007-07-17  80->256
    char outdir[256];      // 2007-07-17  80->256
    char info_file[256]; // 2007-07-17  80->256
    char vex_file[256]; // 2007-07-17  80->256
    char sid[4];
    char dflt_pass[4];
    // for mode2 only run parameres are as follows
    int numch;
    int ntrack;
    int adbit;
    int fanout;
    int parity;
    int nrzm;
    int vlba;
    int ibit[32];
    int tot_track;
	int k5numch; 
	int e2bit;
	int vdifch;
	int chtable[32];   // channel table (contents are thread#)
    //int file_ch_table[4][4]; // channel table for created K5/VSSP file
    int k5nfile;       // 2016-08-08
	int nbitstream;
    int fmode;   // conversion out file mode 1: 1ch mode   2: 4ch mode    3: 4x4ch mode   4:  2x4ch mode
    double fmhz;
    double fsample_orig;  // to keep fsample get or estimated from VDIF data 2019-11-10
    int flag_compulsory;
    int naming_type;   //0: vdifname+'.m5b'(default)  1: vdifname(remove extention)+'.m5b'
    int flag_ch_alloc;
    int pickup_tid;   // pick up thread ID  (-1 means default (1st encountered ID)
    int set_tid;  //setup thread # for output VDIF (default is 0)
    int flag_1sec_adjust;
} runp_t;

typedef struct vdifconv3     // parameters for functions in vdifsch2km
{
    int iwVDIF[32];   // support max 32 thread   
    int64_t iwVDIF64;   // input 64 bit data   
    //int iwK5;
    //int iwK5n[4];
    int *owVDIF;  //   owVDIF[size]    size=thread=numch   
    int64_t *owVDIF64;   // output 64 bit data   
    int pickup_ch;
    int *conv2bit;   // pointer to 2bit decoding table
    int o_adbit;     // AD bits in VDIF out
    int o_numch;     // # of channels in the thread of VDIF out
    int i_numthread;     // # of threads in input VDIF
    int vdifadbit;   //  AD bits in VDIF input
    //int k5nfile;       // 2016-08-08
    int nbit;      // already filled out bits in owVDIF(starting at 0 and < 32)
    int chmax_vdif;
    int bs_vdif;   // bits/stream in VDIF input  2016-08-08  
    //int chtable[16];
    int chtable[32];    // support max 32ch
    int nthsample;
    //double k5fsample;
    double fsample;     // sampling frequency
    int pup_thread_id;   // pick up thread ID 
    int numthread;      // # of tghreads in VDIF out
    int nframesec;      // # of frames/sec
} vdifconv3_t;

#ifndef _VLBI_DATA_TYPE_H_


#define VSSP_FORMAT     0x1          // VSSP    format 
#define VSSP32_FORMAT   0x2          // VSSP32  format 
#define VSSP64_FORMAT   0x3          // VSSP64  format 
#define ADS_FORMAT      0x4          // ADS     format 
#define OCTAD_FORMAT    0x5          // OCTAD   format 
#define VDIF_FORMAT     0x6          // VDIF    format 
#define MK5_FORMAT      0x7          // MARK5   format 
#define M5B_FORMAT      0x8          // MARK5B  format 
#define RAW_FORMAT      0x9          // RAW     format (=ADS and OCTAD)
#define OTHER_FORMAT    0x0          // other   format 


#define ADS_DBBC_MODE  0x1       // ADS DBBC mode
#define ADS_1CH_MODE   0x2       // ADS 1ch/port mode

#define VDIF_KASHIMA_SYNC 0xACABFEEDUL    // sync pattern for VDIF Kashima format
#define VDIF_KASHIMA_FORMAT 0x1     // VDIF legacy format mode (size is 4x8 bytes)
#define VDIF_LEGACY_MODE 0x1     // VDIF legacy format mode (size is 4x8 bytes)

#define VDT_CHMAX 16            // this should be less than equal to CHMAX in "cor_head_new.h" and/or "libcor_new.h"

typedef struct sampling_info  // sampling information
{
   int sfkHz;   //sampling frequency (kHz)
   int adbit;   // AD bits
   int numch;   // # of CH
   int filter;  // LPF filter frequency(MHz) for K5/VSSP32 (0: through)
} sampling_info_t;

typedef struct file_access  // file accesss information
{
    char fname[128];    // file name
    FILE *fi;           // stream pointer
    int flag_open;      // file open flag
    int flag_malloc;    // ibuf malloc flag   1: malloced  0: not malloced
    int *ibuf;          // pointer to read buffer 
    int bufsize;        // size of ibuf
    int nread;      // used for file I/O
    unsigned short *sbuf;
    unsigned char *cbuf;
    int64_t *lbuf;  
    int64_t size;   // file size (bytes)
    int64_t pos;   // current pointer in a file
    int64_t hpos;  // nearest headder start position
    int64_t p1st;  // 1st valid header position (pointer) usually 0
} file_access_t;

typedef struct data_time  // data time information
{
    /* common */
    int year;      // 4 digit year
    int ddd;       // day of year
    int hh;        // hour
    int mm;        // minute
    int ss;        // sec in minute
    int msec;      // msec
    int msec10;      // msec*10 (for Mark5B)
    int second;    // sec in day
    double dtime;  // time in day (ddd+second/86400.0)
    
    /* for VDIF */
	int sec_fm_ref;   // Seconds from reference epoch
	int ref_epo;      // Reference epoch for second count.  
    int month;     // month
    int mday;      // day of month
	int mjd;       // MJD
	double fday;   // fractional part of day
} data_time_t;

typedef struct VDIF_HEADER     // vdif file header structure
{
    sampling_info_t si;
    data_time_t t; // time information converted to YYYY Month Day hour minute second
    unsigned int *ibuf;  // for 1 Frame data including header
	unsigned int *head;   // for header 
	unsigned int *idata;  // for data field 
    int nbuf;          // @# of 32bit words in a VDIF frame including header
	int ndata;         //   # of 32bit data words in a VDIF frame excepting header
    int flag_ibuf_malloc;
	int sec_fm_ref;   // Seconds from reference epoch
	int legacy;       // Legacy mode  0: standard 32 byte VDIF data frame header
	                  //              1: legacy header-length' mode ; words 4-7 omitted from header
	int invalid;      // Invalid data   0: valid  1: invalid
	int frame;        // Data frame # within second, starting at zero; must be integral number od data drames per second
	int ref_epo;      // Reference epoch for second count.  This contains the 6-month period in which the VDIF
	                  // clock was set, with an origin of at 00UTC 1 Jan 2000, such that '0' corresponds to the
	                  // first 6 months of 2000, '1' corresponds to the 6 months starting 00h 1 July 2000, etc. 
	                  // This field rolls over to 0 again at the 6-month period starting 00h 1 Jan 2032.
	int frame_len64;    // Data fram length (including header) in units of 8 bytes (must be multiple of 8 bytes, 
	                  // max is 2^27 bytes)
	int log2_chs;     // log_2(#channels in data array). #chans must be power of 2
	int ver;           // VDIF version number
	unsigned short station_id;   //  station ID (either A2 of unsigned 16-bit integer)
	char sid[3];      // station ID (A2)
	int thread_id;    // Thread ID (0-1023)
    int thread_target; // target thread ID to pcik up (-VE means don't care about thread ID)
	int bpsampl;      // bits per sample (max 32). If complex type, this is set according to the #bits in each complex-
	                  // component
	int d_type;       // data type  0: real data   1: complex data
	int edv;          // Extended data version
	int w1;           // extented data word#1 (3 bytes are valid)
	int w2;           // extended data word#2
	int w3;           // extended data word#3
	int w4;           // extended data word#4
	// below is extended format of NICT's
	int uflag;        // unit of sampling rate 0: kHz,  1: MHz
	int srate;        // sampling rate
	int sync_block;   // sync block  = w2
	char das[9];      // DAS/station name1

} VDIF_HEADER_T;


typedef struct M5B_HEADER             // Mark-5b header type
{
    sampling_info_t si;
    data_time_t t;    // time information
    unsigned int ibuf[2504];  // for 1 Frame data 2504 X 32 bits
    //unsigned int *ibuf;  // for 1 Frame data 2504 X 32 bits
	unsigned int *head;   // for header 4 X 32 bits
	unsigned int *idata;  // for data field 2500 X 32 bits
    //int nbuf=2504;   // # of 32bit words in a Mark5B frame
	//int ndata=2500;  // # of 32bit data words in a Mark5B frame 
    int nbuf;          // # of 32bit words in a Mark5B frame = 2504
	int ndata;         // # of 32bit data words in a Mark5B frame = 2500
	unsigned short uid, icrc16;
	unsigned int iw;
	int itime[6];        // time at begin of frame -1,3digits MJD, hour, minute, sec, msec10
	int iok;
	int frame;        // Data frame # within second, starting at zero; must be integral number od data drames per second
	//int nframe;
	int tvg;
	double fsample;  // sampling frequency
	double p_frame; // frame period
    int ntrack;    // bit# per word  = # of Mark5 tracks  
    int vlba;      // =1: VLBA format  =0; Mark-IV format 
    //int bitsframe; // number of bits per frame (without parity)
} M5B_HEADER_T;

#endif

int convVDIF2VDIF2(runp_t *runp);
void oname_make(runp_t *rp);
//void m5b_hd_make(int *head, int mjd, int seconds, int msec10,
//				 int uid, int tvgflag, int nframe);
int crcc(unsigned char* idata, int len, int mask, int cycl) ;
void m5b_time_update(M5B_HEADER_T *m5b, int mjd1st, int seconds_init, int sec_count);
void make_vdif_header2(vdif_file_t *vd, u_int32_t *head, int frame_byte, int nthread, int numch);
void update_vdif_header2(int secfmepoch, int frame, u_int32_t *head);
void conv_vdifsch2m(vdifconv3_t *ct);
int vdif_read_frame_multi(vdif_file_t *vd);
int get_num_thread_vdif(char *vdif_name, int64_t skip_bytes, vdif_thread_info_t *thd); // 2019-12-11

